from __future__ import annotations

import argparse
import logging
from dataclasses import dataclass

import pwndbg
import pwndbg.aglib.kernel.buddydump
import pwndbg.aglib.kernel.symbol
import pwndbg.aglib.memory
import pwndbg.aglib.symbol
import pwndbg.commands
import pwndbg.dbg_mod
from pwndbg.aglib import kernel
from pwndbg.aglib.kernel import per_cpu
from pwndbg.aglib.kernel.macros import for_each_entry
from pwndbg.commands import CommandCategory
from pwndbg.lib.exception import IndentContextManager

log = logging.getLogger(__name__)


MAX_PG_FREE_LIST_STR_RESULT_CNT = 0x10
MAX_PG_FREE_LIST_CNT = 0x1000
NONE_TUPLE = (None, None)
# https://elixir.bootlin.com/linux/v6.13.12/source/include/linux/mmzone.h#L52
MIGRATE_PCPTYPES = 3


@dataclass
class ParsedBuddyArgs:
    # stores the input options
    zone: str | None
    order: int | None
    mtype: str | None
    cpu: int | None
    find: int | None


@dataclass
class CurrentBuddyParams:
    # stores the current properties of the freelist being/to be traversed
    # this is so that values can be cleanly passed around
    sections: list[tuple[str | None, str | None]]
    indent: IndentContextManager
    node: int | None
    zone: pwndbg.dbg_mod.Value | None
    order: int
    mtype: str
    freelists: pwndbg.dbg_mod.Value | None
    freelist: pwndbg.dbg_mod.Value | None
    nr_types: int | None
    found: bool


def cpu_limitcheck(cpu: str | None):
    if cpu is None:
        return None
    nr_cpus = pwndbg.aglib.kernel.nproc()
    if cpu.isdigit() and int(cpu) < nr_cpus:
        return int(cpu)
    raise argparse.ArgumentTypeError(
        f"The --cpu option takes in a number less than nr_cpu_ids ({nr_cpus})."
    )


parser = argparse.ArgumentParser(
    description="Displays metadata and freelists of the buddy allocator."
)
parser.add_argument(
    "-z",
    "--zone",
    type=str,
    dest="zone",
    default=None,
    help="Displays/searches lists only in the specified zone.",
)
parser.add_argument(
    "-o",
    "--order",
    type=int,
    dest="order",
    help="Displays/searches lists only with the specified order.",
)
parser.add_argument(
    "-m",
    "--mtype",
    type=str,
    dest="mtype",
    default=None,
    help="Displays/searches lists only with the specified mtype.",
)
parser.add_argument(
    "-p",
    "--pcp-only",
    action="store_true",
    dest="pcp_only",
    default=False,
    help="Displays/searches only PCP lists.",
)
parser.add_argument(
    "-c", "--cpu", type=cpu_limitcheck, dest="cpu", default=None, help="CPU nr for searching PCP."
)
parser.add_argument("-n", "--node", type=int, dest="node", default=None, help="")
parser.add_argument(
    "-f",
    "--find",
    type=int,
    dest="find",
    default=None,
    help="The address to find in page free lists.",
)


def static_str_arr(name: str) -> list[str]:
    _arr = pwndbg.aglib.symbol.lookup_symbol(name)
    if not _arr:
        return []
    arr = _arr.dereference()
    return [arr[i].string() for i in range(arr.type.array_len)]


def check_find(counter: int, physmap_addr: int, pba: ParsedBuddyArgs, cbp: CurrentBuddyParams):
    if counter < MAX_PG_FREE_LIST_STR_RESULT_CNT and pba.find is None:
        return True
    if pba.find is None:
        return False
    start = physmap_addr
    end = physmap_addr + pwndbg.aglib.kernel.page_size() * (1 << cbp.order)
    return pba.find >= start and pba.find < end


def traverse_pglist(
    pba: ParsedBuddyArgs, cbp: CurrentBuddyParams
) -> tuple[list[tuple[int, str]] | None, int, list[str] | None]:
    freelist = cbp.freelist
    if freelist is None or int(freelist["next"]) == 0:
        return None, 0, None
    indent = cbp.indent
    seen_pages = set()
    results = []
    counter = 0
    msgs = []
    for e in for_each_entry(freelist, "struct page", "lru"):
        page = int(e)
        phys_addr = pwndbg.aglib.kernel.page_to_phys(page)
        virt_addr = pwndbg.aglib.kernel.page_to_virt(page)
        if check_find(counter, virt_addr, pba, cbp):
            results.append(
                (
                    counter,
                    f"{indent.addr_hex(virt_addr)} [page: {indent.aux_hex(page)}, phys: {indent.aux_hex(phys_addr)}]",
                )
            )
            cbp.found = True
        if counter == MAX_PG_FREE_LIST_STR_RESULT_CNT:
            msgs.append(f"{indent.prefix('... (truncated)')}")
            msgs.append(
                f"This doubly linked list reached size {indent.aux_hex(MAX_PG_FREE_LIST_STR_RESULT_CNT)}"
            )
        counter += 1
        if page in seen_pages:
            msgs.append(f"Cyclic doubly linked list detected: {results[-1]}")
            break
        seen_pages.add(page)
        if counter == MAX_PG_FREE_LIST_CNT:
            msgs.append(
                f"This doubly link list exceeds size {indent.aux_hex(MAX_PG_FREE_LIST_CNT)}"
            )
            break
    return results, counter, msgs


def print_section(section: tuple[str | None, str | None], indent: IndentContextManager):
    prefix, desc = section
    if prefix is not None:
        title = indent.prefix(prefix)
        if desc is not None:
            title = f"{title} ({desc}):"
        indent.print(title)


def print_pglist(pba: ParsedBuddyArgs, cbp: CurrentBuddyParams):
    sections, indent = cbp.sections, cbp.indent
    if len(sections) != 3:
        log.warning(f"The number ({len(sections)}) of sections is not 2!")
        return
    results, counter, msgs = traverse_pglist(pba, cbp)
    if not results or len(results) == 0 or counter == 0:
        return
    print_section(sections[0], indent)
    sections[0] = NONE_TUPLE  # so that the header info is not reprinted
    with indent:
        print_section(sections[1], indent)
        sections[1] = NONE_TUPLE
        with indent:
            print_section(sections[2], indent)
            sections[2] = NONE_TUPLE
            with indent:
                indent.print(
                    f"- {indent.prefix(cbp.mtype)} (contains {indent.aux_hex(counter)} elements)"
                )
                with indent:
                    for i, result in results:
                        indent.print(indent.prefix(f"[0x{i:02x}] ") + result)
                    if msgs is not None:
                        for msg in msgs:
                            indent.print(msg)
                        print()


def print_mtypes(pba: ParsedBuddyArgs, cbp: CurrentBuddyParams):
    freelists, nr_types = cbp.freelists, cbp.nr_types
    mtypes = pwndbg.aglib.kernel.symbol.migratetype_names()
    if nr_types is None:
        nr_types = len(mtypes)
    for i in range(nr_types):
        cbp.mtype = mtypes[i].lower()
        if pba.mtype is not None and cbp.mtype != pba.mtype:
            continue
        if freelists:
            cbp.freelist = freelists[i]
            print_pglist(pba, cbp)


def print_pcp_set(pba: ParsedBuddyArgs, cbp: CurrentBuddyParams):
    assert cbp.zone
    pcp = None
    pcp_lists = None
    if cbp.zone.type.has_field("per_cpu_pageset"):
        pcp = per_cpu(cbp.zone["per_cpu_pageset"], pba.cpu)
        pcp_lists = pcp["lists"]
    elif cbp.zone.type.has_field("pageset"):
        pcp = per_cpu(cbp.zone["pageset"], pba.cpu)
        pcp_lists = pcp["pcp"]["lists"]
    cbp.sections[1] = (f"[cpu #{pba.cpu}] per_cpu_pageset", None)
    if pcp is None or pcp_lists is None:
        log.warning("cannot find pcplist")
        return
    nr_pcp_lists = pwndbg.aglib.kernel.symbol.npcplist()
    for i in range(0, nr_pcp_lists, MIGRATE_PCPTYPES):
        # https://elixir.bootlin.com/linux/v6.13.12/source/include/linux/mmzone.h#L660
        order = i // MIGRATE_PCPTYPES
        if pba.order is not None and pba.order != order:
            continue
        cbp.freelists = pcp_lists[order * MIGRATE_PCPTYPES].address
        cbp.nr_types = MIGRATE_PCPTYPES
        if order == 4:
            # https://elixir.bootlin.com/linux/v6.13/source/arch/x86/include/asm/page_types.h#L20
            order = 21 - 12  # HPAGE_SHIFT - PAGE_SHIFT
            cbp.nr_types = nr_pcp_lists % MIGRATE_PCPTYPES
        cbp.sections[2] = (
            f"order {order}",
            f"size: {cbp.indent.aux_hex(pwndbg.aglib.kernel.page_size() * (1 << order))}",
        )
        cbp.order = order
        print_mtypes(pba, cbp)


def print_free_area(pba: ParsedBuddyArgs, cbp: CurrentBuddyParams):
    assert cbp.zone
    free_area = cbp.zone["free_area"]
    cbp.sections[1] = ("free_area", None)
    for order in range(free_area.type.array_len):
        if pba.order is not None and pba.order != order:
            continue
        cbp.freelists = free_area[order]["free_list"]
        nr_free = int(free_area[order]["nr_free"])
        size = pwndbg.aglib.kernel.page_size() * (1 << order)
        cbp.sections[2] = (
            f"order {order}",
            f"nr_free: {cbp.indent.aux_hex(nr_free)}, size: {cbp.indent.aux_hex(size)}",
        )
        cbp.order = order
        print_mtypes(pba, cbp)


def print_zones(pba: ParsedBuddyArgs, cbp: CurrentBuddyParams, zones, pcp_only):
    target_cpu = pba.cpu
    for i in range(pwndbg.aglib.kernel.symbol.nzones()):
        cbp.zone = zones[i]
        name = zones[i]["name"].string()
        if pba.zone is not None and pba.zone != name:
            continue
        cbp.sections[0] = (f"[node #{cbp.node}] zone {name}", None)
        for cpu in range(pwndbg.aglib.kernel.nproc()):
            if target_cpu is not None and target_cpu != cpu:
                continue
            pba.cpu = cpu
            print_pcp_set(pba, cbp)
        if not pcp_only:
            print_free_area(pba, cbp)


"""
Based off https://github.com/bata24/gef and https://elixir.bootlin.com/linux/v6.13/source

Simplified visualization from bata24/gef:

+-node_data[MAX_NUMNODES]-+
| *pglist_data (node 0)   |--+
| *pglist_data (node 1)   |  |
| *pglist_data (node 2)   |  |
| ...                     |  |
+-------------------------+  |
                             |
+----------------------------+
|
v
+-pglist_data------------------------------+
| node_zones[MAX_NR_ZONES]                 |
|   +-node_zones[0]----------------------+ |   +--->+-per_cpu_pages--------+
|   |  ...                               | |   |    | ...                  |
|   |  per_cpu_pageset                   |-----+    | lists[NR_PCP_LISTS]  |    +-page-----+
|   |  ...                               | |        |   +-lists[0]-------+ |    | flags    |
|   |  name                              | |        |   | next           |----->| lru.next |->..."
|   |  ...                               | |        |   | prev           | |    | lru.prev |
|   |  free_area[MAX_ORDER]              | |        |   +-lists[1]-------+ |    | ...      |
|   |    +-free_area[0]----------------+ | |        |   | ...            | |    +----------+
|   |    | free_list[MIGRATE_TYPES]    | | |        |   +----------------+ |
|   |    |   +-free_list[0]----------+ | | |        +----------------------+
|   |    |   | next                  |---------+
|   |    |   | prev                  | | | |   |
|   |    |   +-free_list[1]----------+ | | |   |    +-page-----+    +-page-----+    +-page-----+
|   |    |   | ...                   | | | |   |    | flags    |    | flags    |    | flags    |
|   |    |   +-----------------------+ | | |   +--->| lru.next |--->| lru.next |--->| lru.next |->..."
|   |    | nr_free                     | | |        | lru.prev |    | lru.prev |    | lru.prev |
|   |    +-free_area[1]----------------+ | |        | ...      |    | ...      |    | ...      |
|   |    | ...                         | | |        +----------+    +----------+    +----------+
|   |    +-----------------------------+ | |
|   +-node_zones[1]----------------------+ |
|   |  ...                               | |
|   +------------------------------------+ |
| ...                                      |
+------------------------------------------+
"""


@pwndbg.commands.Command(parser, category=CommandCategory.KERNEL)
@pwndbg.commands.OnlyWhenQemuKernel
@pwndbg.commands.OnlyWithKernelSymbols
@pwndbg.commands.OnlyWhenPagingEnabled
def buddydump(
    zone: str, pcp_only: bool, order: int, mtype: str, cpu: int, node: int, find: int
) -> None:
    node_data = pwndbg.aglib.kernel.node_data()
    if not node_data:
        log.warning("WARNING: Symbol 'node_data' not found")
        return
    pwndbg.aglib.kernel.buddydump.recover_buddydump_typeinfo()
    if not pwndbg.aglib.kernel.has_debug_info():
        node_data = pwndbg.aglib.memory.get_typed_pointer("node_data_t", node_data)
    pba = ParsedBuddyArgs(zone, order, mtype.lower() if mtype is not None else None, cpu, find)
    cbp = CurrentBuddyParams(
        [NONE_TUPLE] * 3, IndentContextManager(), None, None, 0, "", None, None, None, False
    )
    for node_idx in range(kernel.num_numa_nodes()):
        if node is not None and node_idx != node:
            continue
        zones = None
        if "CONFIG_NUMA" in pwndbg.aglib.kernel.kconfig():
            # only display one node per invocation is probably sufficient under most use cases
            zones = node_data.dereference()[node_idx]["node_zones"]
        else:
            zones = node_data["node_zones"]
        cbp.node = node_idx
        print_zones(pba, cbp, zones, pcp_only)
    if not cbp.found:
        log.warning("No free pages with specified filters found.")



from __future__ import annotations

import re
from asyncio import CancelledError
from collections.abc import Callable
from collections.abc import Coroutine
from collections.abc import Generator
from collections.abc import Iterator
from collections.abc import Sequence
from contextlib import contextmanager
from contextlib import nullcontext
from os import environ
from pathlib import Path
from random import randint
from typing import Any
from typing import Literal
from typing import TypeVar

import gdb
import gdb.types
from typing_extensions import override

import pwndbg
import pwndbg.color.message as message
import pwndbg.dbg_mod
import pwndbg.gdblib
import pwndbg.gdblib.events
import pwndbg.lib.cache
import pwndbg.lib.memory
import pwndbg.lib.path
from pwndbg.dbg_mod import EventHandlerPriority
from pwndbg.dbg_mod import EventType
from pwndbg.dbg_mod import selection
from pwndbg.gdblib import load_gdblib
from pwndbg.lib.arch import ArchAttribute
from pwndbg.lib.arch import ArchDefinition
from pwndbg.lib.arch import Platform
from pwndbg.lib.siginfo import SigInfo
from pwndbg.lib.siginfo import SigInfoKill
from pwndbg.lib.siginfo import SigInfoRt
from pwndbg.lib.siginfo import SigInfoSigChld
from pwndbg.lib.siginfo import SigInfoSigFault
from pwndbg.lib.siginfo import SigInfoSigPoll
from pwndbg.lib.siginfo import SigInfoSigSys
from pwndbg.lib.siginfo import SigInfoSigVal
from pwndbg.lib.siginfo import SigInfoTimer

T = TypeVar("T")


# List of supported architectures that GDB recognizes
# These strings to converted to the Pwndbg-specific name for the architecture
gdb_architecture_name_fixup_list = (
    "x86-64",
    "i386",
    "i8086",
    "aarch64",
    "mips",
    "rs6000",
    "powerpc",
    "sparc",
    "arm",
    "iwmmxt",
    "iwmmxt2",
    "xscale",
    "riscv:rv32",
    "riscv:rv64",
    "riscv",
    "loongarch64",
    "s390:64-bit",
)

# `show architecture` returns a string like "mips:isa32r5"
gdb_mips_to_arch_attribute_map = {
    "mips5": ArchAttribute.MIPS_ISA_5,
    "micromips": ArchAttribute.MIPS_ISA_MICRO,
    "isa32": ArchAttribute.MIPS_ISA_32,
    "isa32r2": ArchAttribute.MIPS_ISA_32R2,
    "isa32r3": ArchAttribute.MIPS_ISA_32R3,
    "isa32r5": ArchAttribute.MIPS_ISA_32R5,
    "isa32r6": ArchAttribute.MIPS_ISA_32R6,
    "isa64": ArchAttribute.MIPS_ISA_64,
    "isa64r2": ArchAttribute.MIPS_ISA_64R2,
    "gs264e": ArchAttribute.MIPS_ISA_64R2,
    "gs464": ArchAttribute.MIPS_ISA_64R2,
    "gs464e": ArchAttribute.MIPS_ISA_64R2,
    "isa64r3": ArchAttribute.MIPS_ISA_64R3,
    "isa64r5": ArchAttribute.MIPS_ISA_64R5,
    "isa64r6": ArchAttribute.MIPS_ISA_64R6,
}


def parse_and_eval(expression: str, global_context: bool) -> gdb.Value:
    """
    Same as `gdb.parse_and_eval`, but only uses `global_context` if it is
    supported by the current version of GDB.

    `global_context` was introduced in GDB 14.
    """
    try:
        return gdb.parse_and_eval(expression, global_context)
    except TypeError:
        return gdb.parse_and_eval(expression)


def _get_frame_stack_variables(frame: gdb.Frame) -> tuple[tuple[int, int, str], ...]:
    try:
        block = frame.block()
    except (gdb.error, RuntimeError):
        # gdb.error: No frame selected (no active inferior)
        # RuntimeError: Frame exists and selected,
        # But no DWARF info, such as in the case of stripped binaries
        return ()

    if not block:
        return ()

    variables = []
    while block:
        for sym in block:
            if not (sym.is_variable or sym.is_argument):
                continue

            try:
                value = sym.value(frame)
                # value.address can be None
                # https://sourceware.org/gdb/current/onlinedocs/gdb.html/Values-From-Inferior.html#Values-From-Inferior:~:text=Variable%3A%20Value%2Eaddress
                # https://sourceware.org/bugzilla/show_bug.cgi?id=33860
                if value.address is None:
                    continue

                addr = int(value.address)
                size = value.type.sizeof
                variables.append((addr, addr + size, sym.name))
            except (gdb.error, AttributeError):
                # FIXME: We should get rid of this try-except in favour of checking the necessary conditions beforehand.
                continue

        block = block.superblock

    return tuple(variables)


class GDBRegisters(pwndbg.dbg_mod.Registers):
    def __init__(self, frame: GDBFrame):
        self.frame = frame

    @override
    def by_name(self, name: str) -> pwndbg.dbg_mod.Value | None:
        try:
            return GDBValue(self.frame.inner.read_register(name))
        except (gdb.error, ValueError):
            # GDB throws an exception if the name is unknown, we just return
            # None when that is the case.
            pass
        return None


class GDBFrame(pwndbg.dbg_mod.Frame):
    def __init__(self, inner: gdb.Frame):
        self.inner = inner
        # caching the hash since calculating it involves some string operations
        self._hash: int | None = None

    @override
    def lookup_symbol(
        self,
        name: str,
        *,
        type: pwndbg.dbg_mod.SymbolLookupType = pwndbg.dbg_mod.SymbolLookupType.ANY,
    ) -> pwndbg.dbg_mod.Value | None:
        from pwndbg.dbg_mod.gdb.symbol import Domain
        from pwndbg.dbg_mod.gdb.symbol import lookup_frame_symbol

        domain = {
            pwndbg.dbg_mod.SymbolLookupType.ANY: Domain.ANY,
            pwndbg.dbg_mod.SymbolLookupType.VARIABLE: Domain.VARIABLE,
            pwndbg.dbg_mod.SymbolLookupType.FUNCTION: Domain.FUNCTION,
        }[type]
        try:
            if (val := lookup_frame_symbol(name, domain=domain)) is not None:
                return GDBValue(val)
        except gdb.error as e:
            raise pwndbg.dbg_mod.Error(e)
        return None

    @override
    def evaluate_expression(
        self, expression: str, lock_scheduler: bool = False
    ) -> pwndbg.dbg_mod.Value:
        from pwndbg.gdblib.scheduler import lock_scheduler as do_lock_scheduler

        with do_lock_scheduler() if lock_scheduler else nullcontext():
            with selection(self.inner, lambda: gdb.selected_frame(), lambda f: f.select()):
                try:
                    value = parse_and_eval(expression, global_context=False)
                except gdb.error as e:
                    raise pwndbg.dbg_mod.Error(e)

        return GDBValue(value)

    @override
    def regs(self) -> pwndbg.dbg_mod.Registers:
        return GDBRegisters(self)

    @override
    def reg_write(self, name: str, val: int) -> bool:
        import pwndbg.aglib

        if name not in pwndbg.aglib.regs:
            return False

        with selection(self.inner, lambda: gdb.selected_frame(), lambda f: f.select()):
            try:
                gdb.execute(f"set ${name} = {val}")
                return True
            except gdb.error as e:
                raise pwndbg.dbg_mod.Error(e)

    @override
    def pc(self) -> int:
        return int(self.inner.pc())

    @override
    def sp(self) -> int:
        # I don't see this exposed anywhere in the API nor in `info frame`, so we will have
        # to read from the inferior. `sp` is resolved by GDB to the architecture-specific stack pointer.
        return int(self.regs().by_name("sp"))

    @override
    def start(self) -> int | None:
        # How is it possible that this isn't in the API?
        # https://sourceware.org/gdb/current/onlinedocs/gdb.html/Frames-In-Python.html#Frames-In-Python
        import pwndbg.aglib

        # We're gonna parse this:
        # e.g. `str(self.inner) == "{stack=0x7fffffffe030,code=0x00007ffff7fe0880,!special}"`
        # See gdb/python/py-frame.c:frapy_str() and gdb/frame.c:frame_id::to_string()
        # They really could just expose .stack() as an API...
        # We could have also chosen to parse `gdb.execute("info frame")`.

        str_id = str(self.inner)
        if "stack=0x" in str_id:
            # GDB for some reason returns one ptr past retaddr
            return (
                int(str_id.partition("stack=")[2].partition(",")[0], 16) - pwndbg.aglib.arch.ptrsize
            )
        # We got "!stack", "stack=<unavailable>", "stack=<sentinel>" or "stack=<outer>".
        return None

    @override
    def parent(self) -> pwndbg.dbg_mod.Frame | None:
        try:
            parent = self.inner.older()
            if parent is not None:
                return GDBFrame(parent)
        except (gdb.error, gdb.MemoryError) as e:
            # We can encounter a `gdb.error: PC not saved` here.
            raise pwndbg.dbg_mod.Error(e)

        return None

    @override
    def child(self) -> pwndbg.dbg_mod.Frame | None:
        try:
            child = self.inner.newer()
            if child is not None:
                return GDBFrame(child)
        except (gdb.error, gdb.MemoryError) as e:
            # We can encounter a `gdb.error: PC not saved` here.
            raise pwndbg.dbg_mod.Error(e)

        return None

    @override
    def sal(self) -> tuple[str, int] | None:
        sal = self.inner.find_sal()  # gdb.Symtab_and_line
        if sal.symtab is None:
            return None

        return sal.symtab.fullname(), sal.line

    @override
    @pwndbg.lib.cache.cache_until("forever")
    def stack_variables(self) -> tuple[tuple[int, int, str], ...]:
        import pwndbg.aglib.qemu

        if pwndbg.aglib.qemu.is_qemu_kernel():
            # Workaround for GDB bug:
            # https://sourceware.org/bugzilla/show_bug.cgi?id=33861
            # https://github.com/pwndbg/pwndbg/issues/3693
            # Unfortunately there is no consistent way to detect this, I've had it
            # happen even on block0 with hex(block.end - block.start) == 0x14a,
            # so we simply don't support getting stack variables on GDB when kernel
            # debugging.
            return ()
        return _get_frame_stack_variables(self.inner)

    @override
    def __eq__(self, rhs: object) -> bool:
        assert isinstance(rhs, GDBFrame), "tried to compare GDBFrame to other type"
        other: GDBFrame = rhs

        return self.inner == other.inner

    @override
    def idx(self) -> int:
        # https://sourceware.org/gdb/current/onlinedocs/gdb.html/Frames-In-Python.html#Frames-In-Python:~:text=Frame%2Elevel
        return self.inner.level()

    def _frame_id(self) -> tuple[int, int, int, int]:
        """
        Returns a tuple which recovers the frame_id struct of the GDB frame object (self.inner).
        We cannot recover only frame_id.user_created_p.

        Format is (stack_addr, code_addr, special_addr, artificial_depth).

        The status fields are returned implicitly in the following:
            stack_addr == -1 means stack_status == FID_STACK_INVALID
            stack_addr == -2 means stack_status == FID_STACK_UNAVAILABLE
            stack_addr == -3 means stack_status == FID_STACK_SENTINEL
            stack_addr == -4 means stack_status == FID_STACK_OUTER
            code_addr == -1 means code_addr_p == 0
            special_addr == -1 means special_addr_p == 0
        Note that artificial_depth == 0 means invalid.
        Note that stack_addr is the start of the stack, not the value of the stack pointer.
        """
        # See gdb/frame-id.h and gdb/frame.c:frame_id::to_string().
        # This obviously isn't guaranteed to be stable, but looking at the blame
        # of to_string() it has been.

        # Here is an example of what is can look like:
        # "{stack=0x7fffffffe030,code=0x00007ffff7fe0880,!special}"
        frame_id_str: str = str(self.inner)

        # Get rid of curly braces and split into individual fields.
        splitted: list[str] = frame_id_str[1:-1].split(",")

        stack_addr: int
        code_addr: int
        special_addr: int
        artificial_addr: int

        match splitted[0]:
            case "!stack":
                stack_addr = -1
            case "stack=<unavailable>":
                stack_addr = -2
            case "stack=<sentinel>":
                stack_addr = -3
            case "stack=<outer>":
                stack_addr = -4
            case _:
                # We have an actual address (len("stack=") == 6)
                stack_addr = int(splitted[0][6:], 16)

        if splitted[1] == "!code":
            code_addr = -1
        else:
            # len("code=") == 5
            code_addr = int(splitted[1][5:], 16)

        if splitted[2] == "!special":
            special_addr = -1
        else:
            # len("special=") == 8
            special_addr = int(splitted[2][8:], 16)

        if len(splitted) == 3:
            artificial_addr = -1
        else:
            # len("artificial=") == 11
            artificial_addr = int(splitted[3][11:], 16)

        return (stack_addr, code_addr, special_addr, artificial_addr)

    @override
    def __hash__(self) -> int:
        if self._hash is not None:
            # The object is immutable, so this works.
            return self._hash

        # GDB implements the equality comparison in gdb/python/py-frame.c:frapy_richcompare()
        # and gdb/frame.c:frame_id::operator==().
        # The frame_id is not exposed to the python API, but we can exploit str(self.inner)
        # to recover it. (though still don't have frame_id_is_next, but oh well..)
        # This is much faster than reading the stack pointer from the inferior, and will give
        # more accurate results as well.
        stack_addr, code_addr, special_addr, artificial_depth = self._frame_id()
        # stack_addr == -1 (FID_STACK_INVALID) acts as NaN, so we have to be unique
        if stack_addr == -1:
            self._hash = randint(0, 1 << 63)
            return self._hash

        # I don't want to make assumptions on the sizes of all these fields so I will just build up a
        # string and use the string's hash. Using "|" for domain separation.
        self._hash = hash(
            f"{stack_addr}|{code_addr}|{special_addr}|{artificial_depth}|"
            # Still using idx() for good measure (maybe it helps with the lack of frame_id_is_next?)
            # Still using number_of_stops_since_birth because I want to guarantee that same frames at
            # different times return different hashes.
            f"{self.idx()}|{pwndbg.dbg_mod.number_of_stops_since_birth}"
        )

        return self._hash


class GDBThread(pwndbg.dbg_mod.Thread):
    def __init__(self, inner: gdb.InferiorThread):
        self.inner = inner

    @override
    @contextmanager
    def bottom_frame(self) -> Iterator[pwndbg.dbg_mod.Frame]:
        with selection(self.inner, lambda: gdb.selected_thread(), lambda t: t.switch()):
            yield GDBFrame(gdb.newest_frame())

    @override
    def ptid(self) -> int | None:
        _, lwpid, _ = self.inner.ptid
        return lwpid

    @override
    def index(self) -> int:
        return self.inner.num

    @override
    def siginfo(self) -> SigInfo | None:
        try:
            gdb_siginfo_expr = pwndbg.dbg.selected_inferior().evaluate_expression("$_siginfo")
            siginfo = SigInfo(
                si_signo=int(gdb_siginfo_expr["si_signo"]),
                si_errno=int(gdb_siginfo_expr["si_errno"]),
                si_code=int(gdb_siginfo_expr["si_code"]),
                kill=SigInfoKill(
                    si_pid=int(gdb_siginfo_expr["_sifields"]["_kill"]["si_pid"]),
                    si_uid=int(gdb_siginfo_expr["_sifields"]["_kill"]["si_uid"]),
                ),
                timer=SigInfoTimer(
                    si_tid=int(gdb_siginfo_expr["_sifields"]["_timer"]["si_tid"]),
                    si_overrun=int(gdb_siginfo_expr["_sifields"]["_timer"]["si_overrun"]),
                    si_sigval=SigInfoSigVal(
                        sival_int=int(
                            gdb_siginfo_expr["_sifields"]["_timer"]["si_sigval"]["sival_int"]
                        ),
                        sival_ptr=int(
                            gdb_siginfo_expr["_sifields"]["_timer"]["si_sigval"]["sival_ptr"]
                        ),
                    ),
                ),
                rt=SigInfoRt(
                    si_pid=int(gdb_siginfo_expr["_sifields"]["_rt"]["si_pid"]),
                    si_uid=int(gdb_siginfo_expr["_sifields"]["_rt"]["si_uid"]),
                    si_sigval=SigInfoSigVal(
                        sival_int=int(
                            gdb_siginfo_expr["_sifields"]["_rt"]["si_sigval"]["sival_int"]
                        ),
                        sival_ptr=int(
                            gdb_siginfo_expr["_sifields"]["_rt"]["si_sigval"]["sival_ptr"]
                        ),
                    ),
                ),
                sigchld=SigInfoSigChld(
                    si_pid=int(gdb_siginfo_expr["_sifields"]["_sigchld"]["si_pid"]),
                    si_uid=int(gdb_siginfo_expr["_sifields"]["_sigchld"]["si_uid"]),
                    si_status=int(gdb_siginfo_expr["_sifields"]["_sigchld"]["si_status"]),
                    si_utime=int(gdb_siginfo_expr["_sifields"]["_sigchld"]["si_utime"]),
                    si_stime=int(gdb_siginfo_expr["_sifields"]["_sigchld"]["si_stime"]),
                ),
                sigfault=SigInfoSigFault(
                    si_addr=int(gdb_siginfo_expr["_sifields"]["_sigfault"]["si_addr"]),
                ),
                sigpoll=SigInfoSigPoll(
                    si_band=int(gdb_siginfo_expr["_sifields"]["_sigpoll"]["si_band"]),
                    si_fd=int(gdb_siginfo_expr["_sifields"]["_sigpoll"]["si_fd"]),
                ),
                sigsys=SigInfoSigSys(
                    call_addr=int(gdb_siginfo_expr["_sifields"]["_sigsys"]["_call_addr"]),
                    syscall=int(gdb_siginfo_expr["_sifields"]["_sigsys"]["_syscall"]),
                    arch=int(gdb_siginfo_expr["_sifields"]["_sigsys"]["_arch"]),
                ),
            )
            return siginfo
        except pwndbg.dbg_mod.Error:
            return None


class GDBMemoryMap(pwndbg.dbg_mod.MemoryMap):
    def __init__(self, qemu: bool, pages: Sequence[pwndbg.lib.memory.Page]):
        super().__init__(pages)
        self.qemu = qemu

    @override
    def is_qemu(self) -> bool:
        return self.qemu


# While this implementation allows breakpoints to be deleted, enabled and
# disabled from inside the code in a stop handler, GDB does not[1]. Aditionally,
# it behaves largely unpredictably when we try to do it. So, in order to allow
# for these things, we defer the operations on the GDB side until we're sure
# we can do them, and do some extra work on the Pwndbg side.
#
# [1]: https://sourceware.org/gdb/current/onlinedocs/gdb.html/Breakpoints-In-Python.html#Breakpoints-In-Python
BPWP_DEFERRED_DELETE: set[GDBStopPoint] = set()
BPWP_DEFERRED_ENABLE: set[GDBStopPoint] = set()
BPWP_DEFERRED_DISABLE: set[GDBStopPoint] = set()
# See pwndbg/dbg_mod/gdb/hooks.py !


class BreakpointAdapter(gdb.Breakpoint):
    stop_handler: Callable[[], bool]

    @override
    def stop(self) -> bool:
        return self.stop_handler()


class FinishpointAdapter(gdb.FinishBreakpoint):
    stop_handler: Callable[[], bool]

    def __init__(self, stop_handler, internal):
        super().__init__(gdb.newest_frame(), internal)
        self.stop_handler = stop_handler

    @override
    def stop(self) -> bool:
        result = self.stop_handler()
        return result


class GDBStopPoint(pwndbg.dbg_mod.StopPoint):
    inner: gdb.Breakpoint
    proc: GDBProcess
    inner_stop: Callable[[], bool] | None

    def __init__(self, inner: gdb.Breakpoint, proc: GDBProcess):
        self.inner = inner
        self.proc = proc
        self.inner_stop = None

    def _stop(self):
        """
        This function implements the same protocol as the GDB stop() function
        and may be slotted in place of the original function in case we need to
        disable or delete a breakpoint or watchpoint during the handling of
        a stop function.
        """
        if self not in BPWP_DEFERRED_DISABLE and self not in BPWP_DEFERRED_DELETE:
            return self.inner_stop()
        return False

    def _clear(self):
        """
        Removes the soft-disable aware handler and restores the original handler,
        if one was installed.
        """
        if self.inner_stop is not None:
            self.inner.stop = self.inner_stop
            self.inner_stop = None

    @override
    def set_enabled(self, enabled: bool) -> None:
        if self.proc.in_bpwp_stop_handler:
            # We're doing this during a stop handle. Change the stop function
            # in the breakpoint for the version that supports soft-disabling of
            # the breakpoint and then soft-disable it.
            self.inner_stop = self.inner.stop
            self.inner.stop = self._stop

            if enabled:
                target = BPWP_DEFERRED_ENABLE
                other = BPWP_DEFERRED_DISABLE
            else:
                target = BPWP_DEFERRED_DISABLE
                other = BPWP_DEFERRED_ENABLE
            if self in other:
                other.remove(self)

            target.add(self)
        else:
            # We're not in the middle of a stop handle, just enable or disable
            # it directly in GDB.
            self.inner.enabled = enabled

    @override
    def remove(self) -> None:
        if self.proc.in_bpwp_stop_handler:
            # Same as in `set_enabled`. We can't actually disable it right away,
            # but we can stop the handle from running and prevent the breakpoint
            # from stopping the program until it actually gets deleted.
            self.inner_stop = self.inner.stop
            self.inner.stop = self._stop
            BPWP_DEFERRED_DELETE.add(self)
        else:
            self.inner.delete()


class GDBProcess(pwndbg.dbg_mod.Process):
    # Operations that change the internal state of GDB are generally not allowed
    # during breakpoint stop handles. Because the Pwndbg Debugger-agnostic API
    # generally does not have this limitation, we keep track of these handles,
    # in order to properly block off or implement operations we support, but
    # that GDB would misbehave doing.
    in_bpwp_stop_handler: bool

    def __init__(self, inner: gdb.Inferior):
        self.inner = inner
        self.in_bpwp_stop_handler = False

    @override
    def threads(self) -> list[pwndbg.dbg_mod.Thread]:
        return [GDBThread(thread) for thread in gdb.selected_inferior().threads()]

    @override
    def pid(self) -> int | None:
        i = gdb.selected_inferior()
        if i is not None:
            return i.pid
        return None

    @override
    def alive(self) -> bool:
        return gdb.selected_thread() is not None

    @override
    def is_core_file(self) -> bool:
        inferior = gdb.selected_inferior()
        return (
            hasattr(inferior, "connection")
            and inferior.connection is not None
            and inferior.connection.type == "core"
        )

    @override
    def stopped_with_signal(self) -> bool:
        return "It stopped with signal " in gdb.execute("info program", to_string=True)

    @override
    def stopped_at_breakpoint(self) -> bool:
        gdb_prog: str = gdb.execute("info program", to_string=True)
        # The first happens when e.g. running start
        # ("It stopped at a breakpoint that has since been deleted.")
        # the second on actual user set breakpoints
        # ("It stopped at breakpoint 2.")
        return "It stopped at a breakpoint " in gdb_prog or "It stopped at breakpoint " in gdb_prog

    @override
    def evaluate_expression(self, expression: str) -> pwndbg.dbg_mod.Value:
        try:
            return GDBValue(parse_and_eval(expression, global_context=True))
        except gdb.error as e:
            raise pwndbg.dbg_mod.Error(e)

    @override
    def vmmap(self) -> pwndbg.dbg_mod.MemoryMap:
        import pwndbg.aglib.qemu
        from pwndbg.aglib.kernel.vmmap import kernel_vmmap
        from pwndbg.aglib.vmmap_custom import get_custom_pages
        from pwndbg.gdblib.vmmap import get_known_maps

        qemu = pwndbg.aglib.qemu.is_old_qemu_user()
        proc_maps = get_known_maps()
        # The `proc_maps` is usually a tuple of Page objects but it can also be:
        #   None    - when /proc/$tid/maps does not exist/is not available
        #   tuple() - when the process has no maps yet which happens only during its very early init
        #             (usually when we attach to a process)
        if proc_maps is not None:
            return GDBMemoryMap(qemu, proc_maps)

        pages: list[pwndbg.lib.memory.Page] = []
        pages.extend(kernel_vmmap())
        pages.extend(get_custom_pages())
        pages.sort()
        return GDBMemoryMap(qemu, pages)

    def _is_memory_readable(self, addr: int) -> bool:
        try:
            gdb.selected_inferior().read_memory(addr, 1)
            return True
        except gdb.error:
            return False

    def _find_memory_last_readable(self, start: int, count: int) -> int:
        end = start + count
        result = -1

        if not self._is_memory_readable(start):
            return result

        while start <= end:
            mid = (start + end + 1) // 2
            if self._is_memory_readable(mid):
                result = mid
                start = mid + 1
            else:
                end = mid - 1

        return result

    @override
    def read_memory(self, address: int, size: int, partial: bool = False) -> bytearray:
        count = max(int(size), 0)
        addr = address

        try:
            result = gdb.selected_inferior().read_memory(addr, count)
            return bytearray(result)
        except gdb.error as e:
            if not partial:
                raise pwndbg.dbg_mod.Error(e)

            if not pwndbg.aglib.remote.is_remote():
                message = str(e)
                match = re.search(r"Memory at address (\w+) unavailable\.", message)
                if match:
                    stop_addr = int(match.group(1), 0)
                else:
                    stop_addr = int(message.split()[-1], 0)

                # Handle case of memory read that wraps around the memory space back to 0, where high memory was readable but memory at 0 was not.
                # Example: 2-byte read at 0xFFFF_FFFF in a 32-bit address space.
                # GDB returns error: "Cannot access memory at address 0x0"
                if stop_addr == 0 and stop_addr < addr:
                    # We could read from the top-portion of memory, but not after wrapping around
                    # Because we are doing a partial read, read until the max address
                    return self.read_memory(addr, pwndbg.aglib.arch.ptrmask - addr + 1)

                if stop_addr > addr:
                    return self.read_memory(addr, stop_addr - addr)
            else:
                # Handle the case of remote debugging, where GDB's remote protocol
                # returns the start address as the failed read address instead of the stop address.
                # This is a limitation in how GDB handles the remote protocol, and while it could
                # be fixed, it currently behaves this way.
                #
                # To work around this, we perform a binary search in the `_find_memory_last_readable` method
                # to find the correct stop address that avoids the failure.
                #
                # For local debugging, this issue does not occur, and we proceed with the normal flow.
                if (stop_addr := self._find_memory_last_readable(addr, count)) > 0:
                    return self.read_memory(addr, stop_addr - addr + 1)

            raise pwndbg.dbg_mod.Error(e)

    @override
    def write_memory(self, address: int, data: bytearray, partial: bool = False) -> int:
        try:
            # Throws an exception if can't access memory
            gdb.selected_inferior().write_memory(address, data)
        except gdb.MemoryError as e:
            if partial:
                raise NotImplementedError("partial writes are currently not supported under gdb")

            raise pwndbg.dbg_mod.Error(e)
        return len(data)

    @override
    def find_in_memory(
        self,
        pattern: bytearray,
        start: int,
        size: int,
        align: int,
        max_matches: int = -1,
        step: int = -1,
    ) -> Generator[int, None, None]:
        if max_matches == 0 or len(pattern) == 0:
            # Nothing to match.
            return

        i = gdb.selected_inferior()
        end = start + size
        limit = None if max_matches < 0 else max_matches
        found_count = 0

        while True:
            # No point in searching if we can't read the memory
            try:
                i.read_memory(start, 1)
            except gdb.MemoryError:
                break

            length = end - start
            if length <= 0:
                break

            try:
                start = i.search_memory(start, length, pattern)
            except gdb.error as e:
                # While remote debugging on an embedded device and searching
                # through a large memory region (~512mb), gdb may return an error similar
                # to `error: Invalid hex digit 116`, even though the search
                # itself is ok. It seems to have to do with a timeout.
                print(f"WARN: gdb.search_memory failed with: {e}")
                if e.args[0].startswith("Invalid hex digit"):
                    print(
                        "WARN: This is possibly related to a timeout. Connection is likely broken."
                    )
                    break
                start = None
                pass

            if start is None:
                break

            # Fix bug: In kernel mode, search_memory may return a negative address,
            # e.g. -1073733344, which supposed to be 0xffffffffc0002120 in kernel.
            start &= 0xFFFFFFFFFFFFFFFF

            # Ignore results that don't match required alignment
            if start & (align - 1):
                start = pwndbg.lib.memory.round_up(start, align)
                continue

            # For some reason, search_memory will return a positive hit
            # when it's unable to read memory.
            try:
                i.read_memory(start, 1)
            except gdb.MemoryError:
                break

            yield start
            found_count += 1
            if limit and found_count >= limit:
                break

            if step > 0:
                start = pwndbg.lib.memory.round_down(start, step) + step
            else:
                if align > 1:
                    start = pwndbg.lib.memory.round_up(start + len(pattern), align)
                else:
                    start += len(pattern)

    @override
    def is_remote(self) -> bool:
        # Example:
        # pwndbg> maintenance print target-stack
        # The current target stack is:
        #   - remote (Remote serial target in gdb-specific protocol)
        #   - exec (Local exec file)
        #   - None (None)
        return "remote" in gdb.execute("maintenance print target-stack", to_string=True)

    @override
    def send_remote(self, packet: str) -> bytes:
        conn = self.inner.connection
        assert isinstance(conn, gdb.RemoteTargetConnection), (
            "Called send_remote() on a local process"
        )
        assert conn.is_valid(), "connection is invalid"

        # NOTE: `send_packet` don't handle reading multiple responses
        try:
            return conn.send_packet(packet) or b""
        except gdb.error as e:
            raise pwndbg.dbg_mod.Error(e)

    @override
    def send_monitor(self, cmd: str) -> str:
        try:
            return gdb.execute(f"monitor {cmd}", to_string=True)
        except gdb.error as e:
            raise pwndbg.dbg_mod.Error(e)

    @override
    def download_remote_file(self, remote_path: str, local_path: str) -> None:
        import pwndbg.aglib.file

        if pwndbg.aglib.file.is_vfile_qemu_user_bug():
            with open(local_path, "wb") as fp:
                try:
                    for data in pwndbg.aglib.file.vfile_readfile(remote_path):
                        fp.write(data)
                    return
                except OSError as e:
                    raise pwndbg.dbg_mod.Error(
                        f"Could not download remote file {remote_path!r}:\nError: {str(e)}"
                    )
        try:
            error = gdb.execute(f'remote get "{remote_path}" "{local_path}"', to_string=True)
        except gdb.error as e:
            error = str(e)

        if error:
            # If the client is configured with set debug remote 1, we need to
            # skip [remote] lines, and not interpret as missing file. Maybe
            # better to search for error strings. A real error will say:
            # "Remote I/O error: No such file or directory"
            real_error = []
            for line in error.splitlines():
                if not line.startswith("[remote]"):
                    real_error.append(line)
            if len(real_error):
                error = "\n".join(real_error)
                raise pwndbg.dbg_mod.Error(
                    f"Could not download remote file {remote_path!r}:\nError: {error}"
                )

    # Note that in GDB this method does not depend on the process at all!
    #
    # From the point-of-view of the GDB implementation, this could very well be
    # implemented as part of Debugger. The issue with that, however, is that the
    # LLDB implementation would have to do some fairly heavy legwork to keep up
    # the appearance that values are independent from any given target.
    #
    # Opting instead to have this method be at this level, although slightly
    # redundant in GDB, saves a ton of work in LLDB.
    @override
    def create_value(
        self, value: int, type: pwndbg.dbg_mod.Type | None = None
    ) -> pwndbg.dbg_mod.Value:
        v = GDBValue(gdb.Value(value))
        if type:
            v = v.cast(type)

        return v

    @override
    def symbol_name_at_address(self, address: int) -> str | None:
        from pwndbg.dbg_mod.gdb.symbol import resolve_addr

        return resolve_addr(address) or None

    @override
    def lookup_symbol(
        self,
        name: str,
        *,
        prefer_static: bool = False,
        type: pwndbg.dbg_mod.SymbolLookupType = pwndbg.dbg_mod.SymbolLookupType.ANY,
        objfile_endswith: str | None = None,
    ) -> pwndbg.dbg_mod.Value | None:
        from pwndbg.dbg_mod.gdb.symbol import Domain
        from pwndbg.dbg_mod.gdb.symbol import lookup_symbol

        domain = {
            pwndbg.dbg_mod.SymbolLookupType.ANY: Domain.ANY,
            pwndbg.dbg_mod.SymbolLookupType.VARIABLE: Domain.VARIABLE,
            pwndbg.dbg_mod.SymbolLookupType.FUNCTION: Domain.FUNCTION,
        }[type]
        try:
            if (
                val := lookup_symbol(
                    name,
                    prefer_static=prefer_static,
                    domain=domain,
                    objfile_endswith=objfile_endswith,
                )
            ) is not None:
                return GDBValue(val)
        except gdb.error as e:
            raise pwndbg.dbg_mod.Error(e)
        return None

    @override
    def get_function_boundaries(self, address: int) -> tuple[int, int] | None:
        block = gdb.block_for_pc(address)

        if block is not None:
            # Final the top-level function that this block resides it
            while block.superblock is not None and block.superblock.function is not None:
                block = block.superblock

            return block.start, block.end

        return None

    @override
    def types_with_name(self, name: str) -> Sequence[pwndbg.dbg_mod.Type]:
        # In GDB, process-level lookups for types are always global.
        #
        # Additionally, the GDB type lookup function only ever returns the first
        # match, so this will always return a list with one element.
        try:
            return [GDBType(gdb.lookup_type(name))]
        except gdb.error:
            return []

    @override
    def arch(self) -> ArchDefinition:
        import pwndbg.aglib.proc
        import pwndbg.aglib.typeinfo

        ptrsize = pwndbg.aglib.typeinfo.ptrsize
        not_exactly_arch = False

        endian: Literal["little", "big"] = None
        if "little" in gdb.execute("show endian", to_string=True).lower():
            endian = "little"
        else:
            endian = "big"

        if pwndbg.aglib.proc.alive():
            arch = gdb.newest_frame().architecture().name()
        else:
            arch = gdb.execute("show architecture", to_string=True).strip()
            not_exactly_arch = True

        arch = arch.lower()

        arch_attributes = []

        if arch.startswith("mips:"):
            isa = arch[5:]

            if (attribute := gdb_mips_to_arch_attribute_map.get(isa)) is not None:
                arch_attributes.append(attribute)
            else:
                message.warn(
                    f"Warning: unable to detect the corrects MIPS variant from GDB architecture '{arch}'\n"
                    "Instruction disassembly may be inaccurate\n"
                    "Please create a GitHub issue including the output of `pi gdb.newest_frame().architecture().name()`"
                )

        # Below, we fix the fetched architecture
        for match in gdb_architecture_name_fixup_list:
            if match in arch:
                # Distinguish between Cortex-M and other ARM
                # When GDB detects correctly Cortex-M processes, it will label them with `arm*-m`, such as armv7e-m
                # However, GDB will sometimes fail to correctly label Cortex-M binaries properly, and says it's simply 'arm'.
                # Internally, GDB still detects the processes as Cortex-M, as it can access .xpsr, but it doesn't
                # appear to expose this in information through any command/API. Since Cortex-M has the .xpsr flags register
                # instead of .cpsr, we will check if it's present.
                # See: https://github.com/pwndbg/pwndbg/issues/2153
                if match == "arm" and (
                    "-m" in arch or pwndbg.aglib.regs.read_reg("xpsr") is not None
                ):
                    match = "armcm"
                elif match.startswith("riscv:"):
                    match = match[6:]
                elif match == "riscv":
                    # If GDB doesn't detect the width, it will just say `riscv`.
                    match = "rv64"
                elif match == "iwmmxt" or match == "iwmmxt2" or match == "xscale":
                    match = "arm"
                elif match == "rs6000":
                    # The RS/6000 architecture is compatible with the PowerPC common
                    match = "powerpc"
                elif match == "s390:64-bit":
                    match = "s390x"
                return ArchDefinition(
                    name=match,  # type: ignore[arg-type]
                    ptrsize=ptrsize,
                    endian=endian,
                    platform=Platform.LINUX,
                    attributes=arch_attributes,
                )

        if not_exactly_arch:
            raise RuntimeError(f"Could not deduce architecture from: {arch}")

        return ArchDefinition(
            name=arch,  # type: ignore[arg-type]
            ptrsize=ptrsize,
            endian=endian,
            platform=Platform.LINUX,
            attributes=arch_attributes,
        )

    @override
    def break_at(
        self,
        location: pwndbg.dbg_mod.BreakpointLocation | pwndbg.dbg_mod.WatchpointLocation,
        stop_handler: Callable[[pwndbg.dbg_mod.StopPoint], bool] | None = None,
        internal: bool = False,
    ) -> pwndbg.dbg_mod.StopPoint:
        # GDB does not support creating new breakpoints in the middle of a
        # breakpoint stop handler[1]. Catch that case and throw an exception.
        #
        # [1]: https://sourceware.org/gdb/current/onlinedocs/gdb.html/Breakpoints-In-Python.html#Breakpoints-In-Python
        if self.in_bpwp_stop_handler:
            raise pwndbg.dbg_mod.Error(
                "Creating new Breakpoints/Watchpoints while in a stop handler is not allowed in GDB"
            )

        if isinstance(location, pwndbg.dbg_mod.BreakpointLocation):
            bp = BreakpointAdapter(
                f"*{location.address:#x}",
                gdb.BP_BREAKPOINT,
                internal=internal,
            )
        elif isinstance(location, pwndbg.dbg_mod.WatchpointLocation):
            if location.watch_read and location.watch_write:
                c = gdb.WP_ACCESS
            elif location.watch_write:
                c = gdb.WP_WRITE
            elif location.watch_read:
                c = gdb.WP_READ

            bp = BreakpointAdapter(
                f"(char[{location.size}])*{location.address}",
                gdb.BP_WATCHPOINT,
                wp_class=c,
                internal=internal,
            )

        if internal:
            bp.silent = True

        sp = GDBStopPoint(bp, self)

        if stop_handler is not None:

            def handler():
                self.in_bpwp_stop_handler = True
                stop = stop_handler(sp)
                self.in_bpwp_stop_handler = False
                return stop
        else:

            def handler():
                return True

        bp.stop_handler = handler

        return sp

    @override
    def trace_ret(self, stop_handler: Callable[[], bool] | None = None, internal: bool = False):
        if stop_handler is None:

            def stop_handler():
                return True

        FinishpointAdapter(stop_handler, internal)

    @override
    def is_linux(self) -> bool:
        # Detect current ABI of client side by 'show osabi'
        #
        # Examples of strings returned by `show osabi`:
        # 'The current OS ABI is "auto" (currently "GNU/Linux").\nThe default OS ABI is "GNU/Linux".\n'
        # 'The current OS ABI is "GNU/Linux".\nThe default OS ABI is "GNU/Linux".\n'
        # 'El actual SO ABI es «auto» (actualmente «GNU/Linux»).\nEl SO ABI predeterminado es «GNU/Linux».\n'
        # 'The current OS ABI is "auto" (currently "none")'
        #
        # As you can see, there might be GDBs with different language versions
        # and so we have to support it there too.
        # Lets assume and hope that `current osabi` is returned in first line in all languages...
        abi = gdb.execute("show osabi", to_string=True).split("\n")[0]

        # Currently we support those osabis:
        # 'GNU/Linux': linux
        # 'none': bare metal

        return "GNU/Linux" in abi

    @override
    def disasm(self, address: int) -> pwndbg.dbg_mod.DisassembledInstruction | None:
        # Currently the type returned by GDB here maps correctly to the type
        # returned by this function, so we don't have to do any extra work.
        #
        # That type is defined in
        # https://sourceware.org/gdb/current/onlinedocs/gdb.html/Architectures-In-Python.html#Architectures-In-Python
        ins: pwndbg.dbg_mod.DisassembledInstruction = (
            gdb.newest_frame().architecture().disassemble(address)[0]
        )
        return ins

    @override
    def module_section_locations(self) -> list[tuple[int, int, str, str]]:
        global pwndbg
        import pwndbg.gdblib.info

        result = []
        for section in pwndbg.gdblib.info.sections():
            result.append(
                (
                    section.start,
                    section.size,
                    section.section,
                    pwndbg.lib.path.clean_path(section.objfile),
                )
            )

        return result

    @override
    def main_module_name(self) -> str | None:
        # Can GDB ever return a different value here from what we'd get with
        # `info files`, give or take a "remote:"?
        if self.alive() and not pwndbg.aglib.qemu.is_qemu_kernel():
            exe = gdb.execute("info proc exe", to_string=True)
            return exe[exe.find("exe = '") + 7 : exe.rfind("'")]
        return gdb.current_progspace().filename

    @override
    def main_module_entry(self) -> int | None:
        import pwndbg.gdblib.info

        for line in pwndbg.gdblib.info.files().splitlines():
            if "Entry point" in line:
                entry_point = int(line.split()[-1], 16)

                # PIE entry points are sometimes reported as an
                # offset from the module base.
                if entry_point < 0x10000:
                    break

                return entry_point

        return None

    @override
    def is_dynamically_linked(self) -> bool:
        out = gdb.execute("info dll", to_string=True)
        return "No shared libraries loaded at this time." not in out

    @override
    def dispatch_execution_controller(
        self, procedure: Callable[[pwndbg.dbg_mod.ExecutionController], Coroutine[Any, Any, None]]
    ):
        # GDB isn't nearly as finnicky as LLDB when it comes to us controlling
        # the execution of the inferior, so we can safely mostly ignore all of
        # the async plumbing and drive the coroutine by just iterating over it.
        #
        # Aditionally, the Debugger-agnostic API allows us enough freedom in how
        # we schedule execution of the controller that running it immediately is
        # perfectly acceptable. So that's what we do.

        coroutine = procedure(EXECUTION_CONTROLLER)
        while True:
            try:
                # We don't need to bother communicating with the coroutine, as
                # it doesn't yield anything we care about.
                coroutine.send(None)
            except (StopIteration, CancelledError):
                # We're done.
                break

    @override
    def add_symbol_file(self, path: str, base: int | None = None) -> None:
        if base is None:
            gdb.execute(f"add-symbol-file {path}", to_string=True)
            return
        gdb.execute(f"add-symbol-file {path} {base}", to_string=True)

    @override
    def remove_symbol_file(self, path: str) -> bool:
        resp: str = gdb.execute(f"remove-symbol-file {path}", to_string=True)
        if "No symbol file found" in resp:
            return False
        return True

    @override
    def runcmd(self, cmd: str) -> str:
        return gdb.execute(cmd, to_string=True)


class GDBExecutionController(pwndbg.dbg_mod.ExecutionController):
    @override
    async def single_step(self):
        # TODO: disable GDB ugly output
        gdb.execute("si")

        # Check if the program stopped because of the step we just took. If it
        # stopped for any other reason, we should propagate a cancellation error
        # to the task and give it a chance to respond.
        if "It stopped after being stepped" not in gdb.execute("info program", to_string=True):
            raise CancelledError()

    @override
    async def cont(self, until: pwndbg.dbg_mod.StopPoint):
        # TODO: disable GDB ugly output
        gdb.execute("continue")

        # Check if the program stopped because of the breakpoint we were given,
        # and, just like for the single step, propagate a cancellation error if
        # it stopped for any other reason.
        assert isinstance(until, GDBStopPoint)
        if f"It stopped at breakpoint {until.inner.number}" not in gdb.execute(
            "info program", to_string=True
        ):
            raise CancelledError()

    @override
    async def cont_selected_thread(self, until: pwndbg.dbg_mod.StopPoint):
        from pwndbg.gdblib.scheduler import lock_scheduler

        with lock_scheduler():
            # TODO: disable GDB ugly output
            gdb.execute("continue")

        # Check if the program stopped because of the breakpoint we were given,
        # and, just like for the single step, propagate a cancellation error if
        # it stopped for any other reason.
        assert isinstance(until, GDBStopPoint)
        if f"It stopped at breakpoint {until.inner.number}" not in gdb.execute(
            "info program", to_string=True
        ):
            raise CancelledError()


# Like in LLDB, we only need a single instance of the execution controller.
EXECUTION_CONTROLLER = GDBExecutionController()


class GDBCommand(gdb.Command):
    def __init__(
        self,
        debugger: GDB,
        name: str,
        handler: Callable[[pwndbg.dbg_mod.Debugger, str, bool], None],
        doc: str | None,
        subcommand_names: list[str] | None,
    ):
        # We can't do `list[str] = []` in the signature because python..
        self.debugger = debugger
        self.handler = handler
        self.__doc__ = doc
        if subcommand_names is None:
            super().__init__(name, gdb.COMMAND_USER, gdb.COMPLETE_EXPRESSION)
        else:
            self.subcommand_names: list[str] = subcommand_names
            super().__init__(name, gdb.COMMAND_USER)

    def invoke(self, args: str, from_tty: bool) -> None:
        self.handler(self.debugger, args, from_tty)

    def complete(self, text: str, word: str | None):
        # https://sourceware.org/gdb/current/onlinedocs/gdb.html/CLI-Commands-In-Python.html#CLI-Commands-In-Python:~:text=Command%2Ecomplete
        # > `text` holds the complete command line up to the cursor’s location
        # > `word` holds the last word of the command line
        # The description is kind of misleading. If you type `slab li ny<TAB>`
        # you will get text="li ny", word="ny".
        # Actually, the function seems to be called twice for some reason, the first invocation having
        # word=None. Why?
        # Since we only support one level of subcommand completion (i.e. we dont support subsubcommand completion),
        # we don't really care about the text and word distinction.
        if word is None or text != word:
            return []
        if text == "":
            # Return all subcommands
            return self.subcommand_names

        # Find all with matching prefix
        return [valid for valid in self.subcommand_names if valid.startswith(text)]


class GDBCommandHandle(pwndbg.dbg_mod.CommandHandle):
    def __init__(self, command: gdb.Command):
        self.command = command

    def remove(self) -> None:
        # GDB doesn't support command removal.
        pass


class GDBType(pwndbg.dbg_mod.Type):
    CODE_MAPPING = {
        gdb.TYPE_CODE_BOOL: pwndbg.dbg_mod.TypeCode.BOOL,
        gdb.TYPE_CODE_INT: pwndbg.dbg_mod.TypeCode.INT,
        gdb.TYPE_CODE_UNION: pwndbg.dbg_mod.TypeCode.UNION,
        gdb.TYPE_CODE_STRUCT: pwndbg.dbg_mod.TypeCode.STRUCT,
        gdb.TYPE_CODE_ENUM: pwndbg.dbg_mod.TypeCode.ENUM,
        gdb.TYPE_CODE_TYPEDEF: pwndbg.dbg_mod.TypeCode.TYPEDEF,
        gdb.TYPE_CODE_PTR: pwndbg.dbg_mod.TypeCode.POINTER,
        gdb.TYPE_CODE_ARRAY: pwndbg.dbg_mod.TypeCode.ARRAY,
        gdb.TYPE_CODE_FUNC: pwndbg.dbg_mod.TypeCode.FUNC,
        # TODO: support `TYPE_CODE_METHOD` differently later?
        gdb.TYPE_CODE_METHOD: pwndbg.dbg_mod.TypeCode.FUNC,
    }

    def __init__(self, inner: gdb.Type):
        self.inner = inner

    @override
    def __eq__(self, rhs: object) -> bool:
        assert isinstance(rhs, GDBType), "tried to compare GDBType to other type"
        other: GDBType = rhs

        return self.inner == other.inner

    @property
    @override
    def name_identifier(self) -> str | None:
        if not self.inner.name:
            return None
        return self.inner.name

    @property
    @override
    def name_to_human_readable(self) -> str:
        if self.inner.name:
            # If named struct/enum/typedef/etc
            return self.inner.name
        return str(self.inner)

    @property
    @override
    def sizeof(self) -> int:
        return self.inner.sizeof

    @property
    @override
    def alignof(self) -> int:
        return self.inner.alignof

    @property
    @override
    def code(self) -> pwndbg.dbg_mod.TypeCode:
        try:
            assert self.inner.code in GDBType.CODE_MAPPING, "missing mapping for type code"
            return GDBType.CODE_MAPPING[self.inner.code]
        except Exception:
            # TODO: log invalid types
            return pwndbg.dbg_mod.TypeCode.INVALID

    @override
    def func_arguments(self) -> list[pwndbg.dbg_mod.Type] | None:
        if self.code != pwndbg.dbg_mod.TypeCode.FUNC:
            raise TypeError("only available for function type")

        # Type without debug info
        # https://github.com/bminor/binutils-gdb/blob/c2dbc2929e87557f8bc030f6f010d67b19f99f12/gdb/gdbtypes.c#L6052-L6072
        is_missing_debug_info = self.inner.name and self.inner.name.endswith(", no debug info>")
        if is_missing_debug_info:
            return None

        args: list[gdb.Field] = self.inner.fields()
        if not args:
            return []
        return [GDBType(arg.type) for arg in args]

    @override
    def fields(self) -> list[pwndbg.dbg_mod.TypeField]:
        return [
            pwndbg.dbg_mod.TypeField(
                field.bitpos if hasattr(field, "bitpos") else 0,
                field.name,
                GDBType(field.type),
                field.parent_type,
                field.enumval if hasattr(field, "enumval") else 0,
                field.artificial,
                field.is_base_class,
                field.bitsize,
            )
            for field in self.inner.fields()
        ]

    @override
    def has_field(self, name: str) -> bool:
        # For GDB, we can do a little better than the default implementation, as
        # it has a specific convenience function that checks for this condition
        # exactly.
        return gdb.types.has_field(self.inner, name)

    @override
    def array(self, count: int) -> pwndbg.dbg_mod.Type:
        # GDB's .array function expects the inclusive upper bound of the array,
        # not the number of elements.
        return GDBType(self.inner.array(count - 1))

    @override
    def pointer(self) -> pwndbg.dbg_mod.Type:
        return GDBType(self.inner.pointer())

    @override
    def strip_typedefs(self) -> pwndbg.dbg_mod.Type:
        return GDBType(self.inner.strip_typedefs())

    @override
    def target(self) -> pwndbg.dbg_mod.Type:
        return GDBType(self.inner.target())

    @override
    def keys(self) -> list[str]:
        return list(self.inner.keys())

    @override
    def offsetof(self, field_name: str) -> int | None:
        # In LLDB this code don't work
        value = pwndbg.dbg.selected_inferior().create_value(0, self.pointer())
        try:
            addr = value[field_name].address
        except pwndbg.dbg_mod.Error:
            # error: `There is no member named field_name`
            return None

        if addr is None:
            raise pwndbg.dbg_mod.Error("bug, this should no happen")

        return int(addr)


class GDBValue(pwndbg.dbg_mod.Value):
    def __init__(self, inner: gdb.Value):
        self.inner = inner

    @property
    @override
    def address(self) -> pwndbg.dbg_mod.Value | None:
        val = self.inner.address
        if val is None:
            return None
        return GDBValue(val)

    @property
    @override
    def is_optimized_out(self) -> bool:
        return self.inner.is_optimized_out

    @property
    @override
    def type(self) -> pwndbg.dbg_mod.Type:
        return GDBType(self.inner.type)

    @override
    def dereference(self) -> pwndbg.dbg_mod.Value:
        if (
            self.type.code == pwndbg.dbg_mod.TypeCode.POINTER
            and self.type.target().code == pwndbg.dbg_mod.TypeCode.FUNC
        ):
            raise pwndbg.dbg_mod.Error("Dereference to function type is not allowed")

        return GDBValue(self.inner.dereference())

    @override
    def string(self) -> str:
        try:
            return self.inner.string()
        except gdb.error as e:
            raise pwndbg.dbg_mod.Error(e)

    @override
    def value_to_human_readable(self) -> str:
        try:
            return str(self.inner)
        except gdb.error as e:
            raise pwndbg.dbg_mod.Error(e)

    @override
    def fetch_lazy(self) -> None:
        self.inner.fetch_lazy()

    @override
    def __int__(self) -> int:
        try:
            return int(self.inner)
        except gdb.error as e:
            raise pwndbg.dbg_mod.Error(e)

    @override
    def cast(self, type: pwndbg.dbg_mod.Type | Any) -> pwndbg.dbg_mod.Value:
        assert isinstance(type, GDBType)
        type: GDBType = type

        if type.code == pwndbg.dbg_mod.TypeCode.FUNC:
            raise pwndbg.dbg_mod.Error("Cast to function type is not allowed, use pointer")

        try:
            return GDBValue(self.inner.cast(type.inner))
        except gdb.error as e:
            # GDB casts can fail.
            raise pwndbg.dbg_mod.Error(e)

    @override
    def __add__(self, rhs: int) -> pwndbg.dbg_mod.Value:
        try:
            return GDBValue(self.inner + rhs)
        except gdb.error as e:
            raise pwndbg.dbg_mod.Error(e)

    @override
    def __sub__(self, rhs: int) -> pwndbg.dbg_mod.Value:
        try:
            return GDBValue(self.inner - rhs)
        except gdb.error as e:
            raise pwndbg.dbg_mod.Error(e)

    @override
    def __getitem__(self, key: str | int) -> pwndbg.dbg_mod.Value:
        if isinstance(key, int) and self.inner.type.strip_typedefs().code == gdb.TYPE_CODE_STRUCT:
            # GDB doesn't normally support indexing fields in a struct by int,
            # so we nudge it a little.
            key = self.inner.type.fields()[key]

        try:
            return GDBValue(self.inner[key])
        except gdb.error as e:
            raise pwndbg.dbg_mod.Error(e)


def _gdb_event_registry_from_event_type(ty: EventType) -> gdb.EventRegistry[Any]:
    """
    Returns the GDB event registry that corresponds to the given event type.
    """
    match ty:
        case EventType.EXIT:
            return gdb.events.exited
        case EventType.CONTINUE:
            return gdb.events.cont
        case EventType.START:
            # Pwndbg installs this one when it loads the GDB event support module.
            #
            # We should never run this function before it gets loaded, but, if this
            # ever changes by mistake, we want the mistake to be caught early, with
            # a clear error.
            assert hasattr(gdb.events, "start"), (
                "gdb.events.start is missing. Did the Pwndbg GDB event code not get loaded?"
            )
            return gdb.events.start
        case EventType.STOP:
            return gdb.events.stop
        case EventType.NEW_MODULE:
            return gdb.events.new_objfile
        case EventType.MEMORY_CHANGED:
            return gdb.events.memory_changed
        case EventType.REGISTER_CHANGED:
            return gdb.events.register_changed
        case EventType.SUSPEND_ALL:
            assert hasattr(gdb.events, "suspend_all"), (
                "gdb.events.suspend_all is missing. Did the Pwndbg GDB event code not get loaded?"
            )
            return gdb.events.suspend_all
        case _:
            raise NotImplementedError(f"unknown event type {ty}")


class GDB(pwndbg.dbg_mod.Debugger):
    def _disable_gdbinit_loading(self) -> tuple[bool, bool]:
        import os

        import psutil

        disable_home_gdbinit = 0
        disable_any_gdbinit = 0
        proc = psutil.Process(os.getpid())
        for arg in proc.cmdline():
            if arg in ("-args", "--args"):
                break
            if arg in ("-nh", "--nh"):
                disable_home_gdbinit += 1
            elif arg in ("-nx", "--nx", "-n", "--n"):
                disable_any_gdbinit += 1

        if disable_any_gdbinit == 0:
            # The `--nx` option is added only in pwndbg-portable mode.
            # This check allows using OLD syntax, eg: `source /path/to/pwndbg/gdbinit.py`, from ~/.gdbinit
            return True, True

        return disable_any_gdbinit >= 2, disable_home_gdbinit >= 1

    def _load_gdbinit(self):
        # Emulate how `gdb` loads `.gdbinit` files (home and local)
        disable_any, disable_home = self._disable_gdbinit_loading()
        if disable_any:
            return

        config_home = environ.get("XDG_CONFIG_HOME", Path("~/.config").expanduser().resolve())
        home_file = (Path(config_home) / "gdb" / "gdbinit").resolve()
        if not home_file.exists():
            home_file = Path("~/.gdbinit").expanduser().resolve()
        local_file = Path("./.gdbinit").resolve()

        def load_source(file_path: str):
            try:
                gdb.execute(f"source {file_path}")
            except gdb.error as e:
                print(e)

        is_home_loaded = False
        if not disable_home and home_file.exists():
            load_source(str(home_file))
            is_home_loaded = True

        disable_local = not gdb.parameter("auto-load local-gdbinit")
        try:
            should_load_local = (
                not disable_local
                and local_file.exists()
                and not (is_home_loaded and home_file.samefile(local_file))
            )
        except PermissionError:
            print(
                message.warn(
                    f"WARNING: Skipped loading {local_file} because you don't have the correct permissions."
                )
            )
            should_load_local = False

        if should_load_local:
            load_source("./.gdbinit")

    @override
    def setup(self):
        import pwnlib.update

        pwnlib.update.disabled = True

        from pwndbg.aglib import load_aglib
        from pwndbg.commands import load_commands

        load_gdblib()
        load_aglib()
        load_commands()

        # Register event hooks.
        # (We can't do them in this file because pwndbg.dbg isn't initialized yet.)
        from pwndbg.dbg_mod.gdb import hooks as hooks

        # Importing `pwndbg.gdblib.prompt` ends up importing code that has the
        # side effect of setting a command up. Because command setup requires
        # `pwndbg.dbg` to already be set, and this module is used as part of the
        # process of setting it, we have to wait, and do the import as part of
        # this method.
        from pwndbg.gdblib import prompt

        prompt.set_prompt()

        pre_commands = """
        set auto-load safe-path /
        set confirm off
        set verbose off
        set pagination off
        set history save on
        set follow-fork-mode child
        set backtrace past-main on
        set step-mode on
        set print pretty on
        set debuginfod enabled on
        handle SIGALRM nostop print nopass
        handle SIGBUS  stop   print nopass
        handle SIGPIPE nostop print nopass
        handle SIGSEGV stop   print nopass
        """

        for line in pre_commands.strip().splitlines():
            gdb.execute(line)

        # This may throw an exception, see pwndbg/pwndbg#27
        try:
            gdb.execute("set disassembly-flavor intel")
        except gdb.error:
            pass

        from pwndbg.gdblib.tui import setup as tui_setup

        tui_setup()

        # Reading Comment file
        from pwndbg.commands import comments

        comments.init()

        from pwndbg.gdblib import config_mod

        config_mod.init_params()

        from pwndbg.dbg_mod.gdb import debug_sym as debug_sym

        self._load_gdbinit()

        # show_hint must be called after loading ~/.gdbinit, this order allow disabling show_hint
        prompt.show_hint()

    @override
    def add_command(
        self,
        name: str,
        handler: Callable[[pwndbg.dbg_mod.Debugger, str, bool], None],
        doc: str | None,
        subcommand_names: list[str] | None = None,
    ) -> pwndbg.dbg_mod.CommandHandle:
        command = GDBCommand(self, name, handler, doc, subcommand_names)
        return GDBCommandHandle(command)

    @override
    def history(self, last: int = 10) -> list[tuple[int, str]]:
        # GDB displays commands in groups of 10. We might want more than that,
        # so we fetch multiple blocks of 10 and assemble them into the final
        # history in a second step.
        parsed_blocks = []
        parsed_lines_count = 0
        parsed_lines_min = None
        parsed_lines_max = None
        parsed_lines_base = None

        while parsed_lines_count < last:
            # Fetch and parse the block we're currently interested in.
            base = f" {parsed_lines_base}" if parsed_lines_base else ""

            lines = gdb.execute(f"show commands{base}", from_tty=False, to_string=True)
            lines = lines.splitlines()

            parsed_lines = []
            for line in lines:
                num_cmd = line.split(maxsplit=1)

                try:
                    number = int(num_cmd[0])
                except ValueError:
                    # In rare cases GDB will output a warning after executing `show commands`
                    # (i.e. "warning: (Internal error: pc 0x0 in read in CU, but not in
                    # symtab.)").
                    return []
                # In rare cases GDB stores a number with no command, and the split()
                # then only returns one element. We can safely ignore these.
                command = num_cmd[1] if len(num_cmd) > 1 else ""

                parsed_lines.append((number, command))

            # We have nothing more to parse if GDB gives us nothing here.
            if len(parsed_lines) == 0:
                break

            # Set the maximum command index we know about. This is simply the
            # last element of the first block.
            if not parsed_lines_max:
                parsed_lines_max = parsed_lines[-1][0]

            # Keep track of the minimum command index we've seen.
            #
            # This is usually the first element in the most recent block, but
            # GDB isn't very clear about whether running commands with
            # `gdb.execute` affects the command history, and what the exact size
            # of the command history is. This means that, at the very end, the
            # first index in the last block might be one greater than the last
            # index in the second-to-last block.
            #
            # Additionally, the value of the first element being greater than
            # the minimum also means that we reached the end of the command
            # history on the last block, can break out of the loop early, and
            # don't even need to bother with this block.
            if parsed_lines_min:
                if parsed_lines[0][0] < parsed_lines_min:
                    parsed_lines_min = parsed_lines[0][0]
                else:
                    break
            else:
                parsed_lines_min = parsed_lines[0][0]

            parsed_blocks.append(parsed_lines)
            parsed_lines_count += len(parsed_lines)

            # If we've just pulled the block with command index 0, we know we
            # can't possibly go back any farther.
            if parsed_lines_base == 0:
                break

            # The way GDB displays the command history is _weird_. The argument
            # we pass to `show commands <arg>` is the index of the 6th element
            # in the block, meaning we'll get a block whose values range from
            # at most <arg> - 5 to at most <arg> + 4, inclusive.
            #
            # Given that we want the first element in this block to the just one
            # past the maximum range of the block returned by the next arguemnt,
            # and that we know the last element in a block is at most <arg> + 4,
            # we can subtract five from its index to land in the right spot.
            parsed_lines_base = max(0, parsed_lines[0][0] - 5)

        # We've got nothing.
        if len(parsed_blocks) == 0:
            return []

        # Sort the elements in the block into the final history array.
        remaining = parsed_lines_max - parsed_lines_min + 1
        plines: list[tuple[int, str]] = [None] * remaining
        while remaining > 0 and len(parsed_blocks) > 0:
            block = parsed_blocks.pop()
            for pline in block:
                index = pline[0] - parsed_lines_min
                if not plines[index]:
                    plines[pline[0] - parsed_lines_min] = pline
                    remaining -= 1

        # If this fails, either some of our assumptions were wrong, or GDB is
        # doing something funky with the output, either way, not good.
        assert remaining == 0, "There are gaps in the command history"

        return plines[-last:]

    @override
    def lex_args(self, command_line: str) -> list[str]:
        return gdb.string_to_argv(command_line)

    @override
    def selected_thread(self) -> pwndbg.dbg_mod.Thread | None:
        thread = gdb.selected_thread()
        if thread:
            return GDBThread(thread)
        return None

    @override
    def selected_frame(self) -> pwndbg.dbg_mod.Frame | None:
        try:
            frame = gdb.selected_frame()
            if frame:
                return GDBFrame(frame)
        except gdb.error:
            pass
        return None

    def commands(self):
        current_pagination = gdb.execute("show pagination", to_string=True)
        current_pagination = current_pagination.split()[-1].rstrip(
            "."
        )  # Take last word and skip period

        gdb.execute("set pagination off")
        command_list = gdb.execute("help all", to_string=True).strip().split("\n")
        existing_commands: set[str] = set()
        for line in command_list:
            line = line.strip()
            # Skip non-command entries
            if (
                not line
                or line.startswith("Command class:")
                or line.startswith("Unclassified commands")
            ):
                continue
            command = line.split()[0]
            existing_commands.add(command)
        gdb.execute(f"set pagination {current_pagination}")  # Restore original setting
        return existing_commands

    @override
    def selected_inferior(self) -> pwndbg.dbg_mod.Process:
        return GDBProcess(gdb.selected_inferior())

    @override
    def is_gdblib_available(self):
        return True

    @override
    def has_event_type(self, ty: EventType) -> bool:
        # Currently GDB supports all event types.
        return True

    @override
    def event_handler(
        self,
        event_type: EventType,
        priority: EventHandlerPriority = EventHandlerPriority.STANDARD,
    ) -> Callable[[Callable[..., None]], Callable[..., None]]:
        import pwndbg.gdblib.events

        event_registry: gdb.EventRegistry[Any] = _gdb_event_registry_from_event_type(event_type)
        return pwndbg.gdblib.events.event_handler_factory(event_registry, event_type.name, priority)

    @override
    @contextmanager
    def ctx_suspend_once(self):
        pwndbg.gdblib.prompt.context_shown = True

    @override
    def suspend_events(self, ty: EventType) -> None:
        pwndbg.gdblib.events.pause(_gdb_event_registry_from_event_type(ty))

    @override
    def resume_events(self, ty: EventType) -> None:
        pwndbg.gdblib.events.unpause(_gdb_event_registry_from_event_type(ty))

    @override
    def set_sysroot(self, sysroot: str) -> bool:
        try:
            gdb.execute(f"set sysroot {sysroot}", from_tty=False)
            # Assume it worked..
            return True
        except gdb.error:
            return False

    @override
    def supports_breakpoint_creation_during_stop_handler(self) -> bool:
        return False

    @override
    def breakpoint_locations(self) -> list[pwndbg.dbg_mod.BreakpointLocation]:
        bps = gdb.breakpoints()
        locations: list[pwndbg.dbg_mod.BreakpointLocation] = []
        for bp in bps:
            if (
                bp.is_valid()
                and bp.enabled
                and bp.type in (gdb.BP_BREAKPOINT, gdb.BP_HARDWARE_BREAKPOINT)
                and bp.visible
            ):
                # GDB 13.1+
                if hasattr(bp, "locations"):
                    for location in bp.locations:
                        locations.append(pwndbg.dbg_mod.BreakpointLocation(location.address))
                else:
                    # Num     Type           Disp Enb Address            What
                    # 1       breakpoint     keep y   0x00007ffff7e90840 in __GI___libc_read at ../sysdeps/unix/sysv/linux/read.c:26
                    bp_locations = gdb.execute(
                        f"info breakpoint {bp.number}", to_string=True
                    ).split("\n")
                    for line in bp_locations:
                        try:
                            address = int(line.split()[4], 16)
                            locations.append(pwndbg.dbg_mod.BreakpointLocation(address))
                        except (IndexError, ValueError):
                            # Ignore lines that don't have an address.
                            pass
        return locations

    @override
    def name(self) -> pwndbg.dbg_mod.DebuggerType:
        return pwndbg.dbg_mod.DebuggerType.GDB

    @override
    def x86_disassembly_flavor(self) -> Literal["att", "intel"]:
        try:
            flavor = gdb.execute("show disassembly-flavor", to_string=True).lower().split('"')[1]
        except gdb.error as e:
            if str(e).find("disassembly-flavor") > -1:
                flavor = "intel"
            else:
                raise pwndbg.dbg_mod.Error(e)

        if flavor != "att" and flavor != "intel":
            raise pwndbg.dbg_mod.Error(f"unrecognized disassembly flavor '{flavor}'")

        literal: Literal["att", "intel"] = flavor
        return literal

    @override
    def string_limit(self) -> int:
        message = gdb.execute("show print elements", from_tty=False, to_string=True)
        message = message.split("\n")[0].split()[-1]
        message = message.strip(".")
        if message == "unlimited":
            return 0
        return int(message)

    @override
    def addrsz(self, address: Any) -> str:
        address = int(address) & pwndbg.aglib.arch.ptrmask
        return f"%#{2 * pwndbg.aglib.arch.ptrsize}x" % address

    @override
    def get_cmd_window_size(self) -> tuple[int | None, int | None]:
        """Get the size of the command window.

        GDB keeps these parameters up to date with the actual window size
        of the command output. This is the full terminal size in CLI mode
        or the size of the cmd window in TUI mode.

        When the window size is set to be unlimited (0), the parameter
        is None.
        """
        width = gdb.parameter("width")
        height = gdb.parameter("height")
        return (
            height if height is None else int(height),
            width if width is None else int(width),
        )

    @override
    @property
    def pre_ctx_lines(self) -> int:
        # GDB most often prints one line
        # as a "0x000055555556fa8f in main ()"-type message
        return 1

    @override
    def set_python_diagnostics(self, enabled: bool) -> None:
        if enabled:
            command = "set python print-stack full"
        else:
            command = "set python print-stack message"

        gdb.execute(command, from_tty=True, to_string=True)

    @override
    def set_convenience_var(self, name: str, value: str, type: str | None) -> None:
        """
        Set a convenience variable which will be accessible with $name in the
        debugger.

        Read the docstring in pwndbg.dbg.set_convenience_var()!!

        Surround this function with try/except.
        """
        if type is not None:
            gdb.execute(f"set ${name} = (({type})({value}))")
        else:
            gdb.execute(f"set ${name} = ({value})")



from __future__ import annotations

import argparse

import pwndbg
import pwndbg.aglib.memory
import pwndbg.commands
import pwndbg.dbg_mod
from pwndbg.commands import CommandCategory


def xor_memory(address, key, count):
    """
    Helper function for xorring memory in gdb
    """
    mem = pwndbg.aglib.memory.read(address, count, partial=True)

    for index, byte in enumerate(mem):
        key_index = index % len(key)
        mem[index] = byte ^ ord(key[key_index])

    return mem


parser = argparse.ArgumentParser(description="XOR `count` bytes at `address` with the key `key`.")
parser.add_argument(
    "address", type=pwndbg.commands.sloppy_gdb_parse, help="The address to start xoring at."
)
parser.add_argument("key", type=str, help="The key to use.")
parser.add_argument("count", type=int, help="The number of bytes to xor.")


@pwndbg.commands.Command(parser, category=CommandCategory.MEMORY)
@pwndbg.commands.OnlyWhenRunning
def xor(address, key, count) -> None:
    try:
        xorred_memory = xor_memory(address, key, count)
        pwndbg.aglib.memory.write(address, xorred_memory)
    except pwndbg.dbg_mod.Error as e:
        print(e)


parser = argparse.ArgumentParser(description="Memfrobs a region of memory (xor with '*').")
parser.add_argument("address", type=int, help="The address to start xoring at.")
parser.add_argument("count", type=int, help="The number of bytes to xor.")


@pwndbg.commands.Command(parser, category=CommandCategory.MEMORY)
@pwndbg.commands.OnlyWhenRunning
def memfrob(address, count):
    return xor(address, "*", count)



"""
Display the kernel ring buffer (dmesg) contents.
This command reads the `printk_ringbuffer` structure, which stores printk messages.
It iterates through the records in the ring buffer to print each record like a dmesg log.

This command supports only the "new" kernel ring buffer implementation that is present in kernel versions 5.10+.
https://git.kernel.org/pub/scm/linux/kernel/git/torvalds/linux.git/commit/?id=d594d8f411d47bf7b583ec3474b11fec348c88bb

This implementation read information from the Linux kernel's `printk_ringbuffer` structure as defined in:
https://github.com/torvalds/linux/blob/19272b37aa4f83ca52bdf9c16d5d81bdd1354494/kernel/printk/printk_ringbuffer.h
"""

from __future__ import annotations

import argparse
import time

import pwndbg.aglib.memory
import pwndbg.aglib.symbol
import pwndbg.aglib.typeinfo
import pwndbg.color.message as message
import pwndbg.commands
import pwndbg.dbg_mod

parser = argparse.ArgumentParser(description="Displays the kernel ring buffer (dmesg) contents.")

parser.add_argument("-T", "--ctime", action="store_true", help="Print human-readable timestamps.")


@pwndbg.commands.Command(
    parser,
    category=pwndbg.commands.CommandCategory.KERNEL,
)
@pwndbg.commands.OnlyWhenQemuKernel
@pwndbg.commands.OnlyWhenPagingEnabled
@pwndbg.commands.OnlyWithKernelDebugInfo
def kdmesg(ctime: bool = False) -> None:
    prb_addr = pwndbg.aglib.symbol.lookup_symbol_addr("printk_rb_static")

    if prb_addr is None:
        print(
            "The printk_rb_static symbol was not found. This may indicate that the kernel is an older version or that the symbol is not available in the current build."
        )
        return

    try:
        # Read printk_ringbuffer structure, which contains a desc_ring and text_data_ring.
        # This struct contains metadata for accessing the ring buffer.
        printk_ringbuffer_type = pwndbg.aglib.memory.get_typed_pointer_value(
            "struct printk_ringbuffer", prb_addr
        )
        desc_ring_addr = printk_ringbuffer_type["desc_ring"].address
        text_data_ring_addr = printk_ringbuffer_type["text_data_ring"].address

        # Read prb_desc_ring structures, which contain metadata about the ring buffer descriptors.
        desc_ring_type = pwndbg.aglib.memory.get_typed_pointer_value(
            "struct prb_desc_ring", desc_ring_addr
        )
        desc_ring_count = 1 << int(desc_ring_type["count_bits"])

        # Find addresses and sizes of descs and infos, neede
        descs = int(desc_ring_type["descs"])
        infos = int(desc_ring_type["infos"])

        # Read prb_data_ring structure, calculate text data size and address.
        # This struct contains the actual text data for the printk messages.
        text_data_ring_type = pwndbg.aglib.memory.get_typed_pointer_value(
            "struct prb_data_ring", text_data_ring_addr
        )
        text_data_sz = 1 << int(text_data_ring_type["size_bits"])
        text_data_addr = int(text_data_ring_type["data"])

        size_long = pwndbg.aglib.typeinfo.lookup_types("long").sizeof
        desc_flags_shift = size_long * 8 - 2
        desc_flags_mask = 3 << desc_flags_shift
        desc_id_mask = ~desc_flags_mask

        # We wish to iterate from tail to head, so we need to find the tail_id and head_id.
        head_id = int(desc_ring_type["head_id"]["counter"])
        tail_id = int(desc_ring_type["tail_id"]["counter"])

        did = int(tail_id)

        prb_desc_size = pwndbg.aglib.typeinfo.load("struct prb_desc").sizeof
        printk_info_size = pwndbg.aglib.typeinfo.load("struct printk_info").sizeof

        # Iterate through each record from tail to head.
        while True:
            ind = did % desc_ring_count
            desc = pwndbg.aglib.memory.get_typed_pointer_value(
                "struct prb_desc", descs + prb_desc_size * ind
            )

            # Skip non-committed or non-finalized records, indicated by the state variable.
            state = 3 & (int(desc["state_var"]["counter"]) >> desc_flags_shift)
            if state != 1 and state != 2:  # desc_committed or desc_finalized
                if did == head_id:
                    break
                did = (did + 1) & desc_id_mask
                continue

            begin = int(desc["text_blk_lpos"]["begin"]) % text_data_sz
            end = int(desc["text_blk_lpos"]["next"]) % text_data_sz

            info = pwndbg.aglib.memory.get_typed_pointer_value(
                "struct printk_info", infos + printk_info_size * ind
            )

            # Read text data from the text_data_ring.
            if begin & 1 == 1:
                text = ""
            else:
                if begin > end:
                    begin = 0

                text_start = begin + size_long
                text_len = int(info["text_len"])

                if end - text_start < text_len:
                    text_len = end - text_start

                text_data = pwndbg.aglib.memory.read(text_data_addr + text_start, text_len)
                text = text_data.decode(encoding="utf8", errors="replace")

            # Format and print the message.
            if ctime:
                tk_core_addr = pwndbg.aglib.symbol.lookup_symbol_addr("tk_core")

                if tk_core_addr is None:
                    print(
                        message.error(
                            "The tk_core symbol was not found. This may indicate that the kernel is an older version or that the symbol is not available in the current build."
                        )
                    )
                    return

                if pwndbg.aglib.typeinfo.load("struct tk_data") is None:
                    print(
                        message.error(
                            "`struct tk_data` is not defined in the current debug symbols."
                        )
                    )
                    return

                tk_core = pwndbg.aglib.memory.get_typed_pointer_value(
                    "struct tk_data", tk_core_addr
                )
                epoch_time = int(tk_core["timekeeper"]["xtime_sec"])

                for line in text.splitlines():
                    print(f"[{time.ctime(int(info['ts_nsec']) / 1e9 + epoch_time)}] {line}")
            else:
                for line in text.splitlines():
                    print(f"[{int(info['ts_nsec']) / 1e9:12.6f}] {line}")

            if did == head_id:
                break
            did = (did + 1) & desc_id_mask

    except pwndbg.dbg_mod.Error as e:
        print(message.error(f"ERROR: {e}"))
        return



from __future__ import annotations

import argparse

import pwndbg.aglib
import pwndbg.aglib.memory
import pwndbg.aglib.stack
import pwndbg.aglib.tls
import pwndbg.auxv
import pwndbg.commands
import pwndbg.commands.telescope
import pwndbg.search
from pwndbg.color import message
from pwndbg.commands import CommandCategory

DEFAULT_NUM_CANARIES_TO_DISPLAY = 1

# Architecture-specific TLS canary offsets
# These offsets are from the TLS base to the canary
# References:
# - x86_64: fs:0x28 (https://elixir.bootlin.com/glibc/glibc-2.41.9000/source/sysdeps/x86_64/nptl/tls.h)
# - i386: gs:0x14 (https://elixir.bootlin.com/glibc/glibc-2.41.9000/source/sysdeps/i386/nptl/tls.h)
# - aarch64: tpidr_el0 + 0x28 (https://elixir.bootlin.com/glibc/glibc-2.41.9000/source/sysdeps/aarch64/nptl/tls.h)
TLS_CANARY_OFFSETS = {
    "x86-64": 0x28,
    "i386": 0x14,
    "aarch64": 0x28,
}


def canary_value() -> tuple[int | None, int | None]:
    """Get the global canary value from AT_RANDOM with its last byte masked (as glibc does)

    Returns:
        tuple: (canary_value, at_random_addr) or (None, None) if not found
    """
    at_random = pwndbg.auxv.get().AT_RANDOM
    if at_random is None:
        return None, None

    global_canary = pwndbg.aglib.memory.read_pointer_width(at_random)

    # masking canary value as canaries on the stack has last byte = 0
    global_canary &= pwndbg.aglib.arch.ptrmask ^ 0xFF

    return global_canary, at_random


def find_tls_canary_addr() -> int | None:
    """Find the address of the canary in the Thread Local Storage (TLS).

    The canary is stored at a fixed offset from the TLS base, which varies by architecture.
    The TLS base can be accessed through architecture-specific registers:
    - x86_64: fs register
    - i386: gs register
    - aarch64: tpidr_el0 register

    Returns:
        int: The virtual address of the canary in TLS, or None if not found/supported
    """
    arch = pwndbg.aglib.arch.name

    # Get TLS base address
    tls_base = (
        pwndbg.aglib.tls.find_address_with_register()
        or pwndbg.aglib.tls.find_address_with_pthread_self()
    )
    if not tls_base:
        return None

    # Get architecture-specific offset
    offset = TLS_CANARY_OFFSETS.get(arch)
    if offset is None:
        return None

    return tls_base + offset


parser = argparse.ArgumentParser(description="Print out the current stack canary.")
parser.add_argument(
    "-a",
    "--all",
    action="store_true",
    help="Print out stack canaries for all threads instead of the current thread only.",
)


@pwndbg.commands.Command(parser, command_name="canary", category=CommandCategory.STACK)
@pwndbg.commands.OnlyWhenRunning
def canary(all) -> None:
    """Display information about the stack canary, including its location in TLS and any copies found on the stack."""
    global_canary, at_random = canary_value()

    if global_canary is None or at_random is None:
        print(message.error("Couldn't find AT_RANDOM - can't display canary."))
        return

    print(message.notice(f"AT_RANDOM  = {at_random:#x} # points to global canary seed value"))

    # Get and display the TLS canary address
    tls_addr = find_tls_canary_addr()
    if tls_addr is not None:
        print(message.notice(f"TLS Canary = {tls_addr:#x} # address where canary is stored"))

        # Verify the value at the TLS address matches our computed canary
        try:
            tls_canary = pwndbg.aglib.memory.read_pointer_width(tls_addr) & (
                pwndbg.aglib.arch.ptrmask ^ 0xFF
            )
            if tls_canary != global_canary:
                print(message.warn("Warning: TLS canary value doesn't match global canary!"))
        except Exception:
            print(message.warn("Warning: Could not read TLS canary value"))
    else:
        print(message.warn("Note: Could not determine TLS canary address for current architecture"))

    print(message.notice(f"Canary     = {global_canary:#x} (may be incorrect on != glibc)"))

    found_canaries = False
    global_canary_packed = pwndbg.aglib.arch.pack(global_canary)
    thread_stacks = pwndbg.aglib.stack.get()
    some_canaries_not_shown = False

    for thread in thread_stacks:
        thread_stack = thread_stacks[thread]

        stack_canaries = list(
            pwndbg.search.search(
                global_canary_packed, start=thread_stack.start, end=thread_stack.end
            )
        )

        if not stack_canaries:
            continue

        found_canaries = True
        num_canaries = len(stack_canaries)
        num_canaries_to_display = num_canaries

        if not all:
            num_canaries_to_display = min(DEFAULT_NUM_CANARIES_TO_DISPLAY, num_canaries)
            if num_canaries_to_display < num_canaries:
                some_canaries_not_shown = True

        if num_canaries > 1:
            print(message.success(f"Thread {thread}: Found valid canaries."))
        else:
            print(message.success(f"Thread {thread}: Found valid canary."))

        for stack_canary in stack_canaries[:num_canaries_to_display]:
            pwndbg.commands.telescope.telescope(address=stack_canary, count=1)

    if found_canaries is False:
        print(message.warn("No canaries found."))

    if some_canaries_not_shown is True:
        print(message.warn("Additional results hidden. Use --all to see them."))



"""
Runs a few useful commands which are available under "info".
"""

from __future__ import annotations

from collections.abc import Iterator
from typing import NamedTuple

import gdb

import pwndbg.lib.cache

# TODO: Add symbol, threads, dll, program


@pwndbg.lib.cache.cache_until("exit")
def proc_mappings() -> str:
    try:
        return gdb.execute("info proc mappings", to_string=True)
    except gdb.error:
        return ""


@pwndbg.lib.cache.cache_until("exit")
def auxv() -> str:
    try:
        return gdb.execute("info auxv", to_string=True)
    except gdb.error:
        return ""


@pwndbg.lib.cache.cache_until("stop")
def files() -> str:
    try:
        return gdb.execute("info files", to_string=True)
    except gdb.error:
        return ""


class Section(NamedTuple):
    objfile: str
    section: str
    start: int
    size: int
    offset: int


def iter_sections() -> Iterator[Section]:
    """
    Parse sections from GDB `maintenance info target-sections`.

    Example output:
    From '/bin/nginx', file type elf64-x86-64:
     [0]      0x00000350->0x00000390 at 0x00000350: .note.gnu.property ALLOC LOAD READONLY DATA HAS_CONTENTS
              Start: 0x558fb29f4350, End: 0x558fb29f4390, Owner token: 0xaaaad5b30c50
     [1]      0x00000390->0x000003b4 at 0x00000390: .note.gnu.build-id ALLOC LOAD READONLY DATA HAS_CONTENTS
              Start: 0x558fb29f4390, End: 0x558fb29f43b4, Owner token: 0xaaaad5b30c50
    """

    # Workaround for GDB bug:
    # Without executing this command first, `maintenance info target-sections`
    # may return incomplete or missing section info. (When using with core-files)
    gdb.execute("info sharedlibrary", to_string=True)

    lines = gdb.execute("maintenance info target-sections", to_string=True)
    current_file = None
    current_section_offset = None
    current_section_name = None

    for line in lines.splitlines():
        line = line.strip()

        # From '/bin/nginx', file type elf64-x86-64:
        if line.startswith("From '"):
            current_file = line[6:].rsplit("'", maxsplit=1)[0]
            continue

        # [0]      0x00000350->0x00000390 at 0x00000350: .note.gnu.property ALLOC LOAD READONLY DATA HAS_CONTENTS
        if line.startswith("["):
            left, right = line.split(": ", maxsplit=1)
            current_section_offset = int(left.split(" at ", maxsplit=1)[1], 16)
            current_section_name = right.split(" ", maxsplit=1)[0]
            continue

        # Start: 0x558fb29f4350, End: 0x558fb29f4390, Owner token: 0xaaaad5b30c50
        if line.startswith("Start:"):
            _, start, _, end, *_ = line.replace(",", "").split(" ")
            start, end = int(start, 16), int(end, 16)

            yield Section(
                objfile=current_file,
                section=current_section_name,
                start=start,
                size=end - start,
                offset=current_section_offset,
            )


@pwndbg.lib.cache.cache_until("stop", "objfile")
def sections() -> tuple[Section, ...]:
    return tuple(iter_sections())



from __future__ import annotations

import argparse

import pwndbg.aglib
import pwndbg.aglib.shellcode
import pwndbg.aglib.vmmap
import pwndbg.color.message as message
import pwndbg.commands
import pwndbg.dbg_mod
import pwndbg.lib.memory
from pwndbg.commands import CommandCategory

parser = argparse.ArgumentParser(
    description="""
Calls the mmap syscall and prints its resulting address.

Note that the mmap syscall may fail for various reasons
(see `man mmap`) and, in case of failure, its return value
will not be a valid pointer.

PROT values: NONE (0), READ (1), WRITE (2), EXEC (4)

MAP values: SHARED (1), PRIVATE (2), SHARED_VALIDATE (3), FIXED (0x10),
            ANONYMOUS (0x20)

Flags and protection values can be either a string containing the names of the
flags or permissions or a single number corresponding to the bitwise OR of the
protection and flag numbers.
""",
)
parser.add_argument(
    "addr", help="Address hint to be given to mmap.", type=pwndbg.commands.sloppy_gdb_parse
)
parser.add_argument(
    "length",
    help="Length of the mapping, in bytes. Needs to be greater than zero.",
    type=int,
)
parser.add_argument(
    "prot",
    help='Prot enum or int as in mmap(2). Eg. "PROT_READ|PROT_EXEC" or 7 (for RWX).',
    type=str,
    nargs="?",
    default="7",
)
parser.add_argument(
    "flags",
    help='Flags enum or int as in mmap(2). Eg. "MAP_PRIVATE|MAP_ANONYMOUS" or 0x22.',
    type=str,
    nargs="?",
    default="0x22",
)
parser.add_argument(
    "fd",
    help="File descriptor of the file to be mapped, or -1 if using MAP_ANONYMOUS.",
    type=int,
    nargs="?",
    default=-1,
)
parser.add_argument(
    "offset",
    help="Offset from the start of the file, in bytes, if using file based mapping.",
    type=int,
    nargs="?",
    default=0,
)
parser.add_argument(
    "--quiet", "-q", help="Disable address validity warnings and hints", action="store_true"
)
parser.add_argument(
    "--force", "-f", help="Force potentially unsafe actions to happen", action="store_true"
)


prot_dict = {
    "PROT_NONE": 0x0,
    "PROT_READ": 0x1,
    "PROT_WRITE": 0x2,
    "PROT_EXEC": 0x4,
}

flag_dict = {
    "MAP_SHARED": 0x1,
    "MAP_PRIVATE": 0x2,
    "MAP_SHARED_VALIDATE": 0x3,
    "MAP_FIXED": 0x10,
    "MAP_ANONYMOUS": 0x20,
}


def prot_str_to_val(protstr):
    """Heuristic to convert PROT_EXEC|PROT_WRITE to integer value."""
    prot_int = 0
    for k, v in prot_dict.items():
        if k in protstr:
            prot_int |= v
    return prot_int


def flag_str_to_val(flagstr):
    """Heuristic to convert MAP_SHARED|MAP_FIXED to integer value."""
    flag_int = 0
    for k, v in flag_dict.items():
        if k in flagstr:
            flag_int |= v
    return flag_int


def parse_str_or_int(val: str | int, parser):
    """
    Try parsing a string with one of the parsers above or by converting it to
    an int, or passes the value through if it is already an integer.
    """
    if isinstance(val, str):
        candidate = parser(val)
        if candidate != 0:
            return candidate
        return int(val, 0)
    if isinstance(val, int):
        return val
    # Getting here is a bug, we shouldn't be seeing other types at all.
    raise TypeError(f"invalid type for value: {type(val)}")


@pwndbg.commands.Command(
    parser,
    category=CommandCategory.MEMORY,
    examples="""
mmap 0x0 4096 PROT_READ|PROT_WRITE|PROT_EXEC MAP_PRIVATE|MAP_ANONYMOUS -1 0
 - Maps a new private+anonymous page with RWX permissions at a location
   decided by the kernel.

mmap 0x0 4096 PROT_READ MAP_PRIVATE 10 0
 - Maps 4096 bytes of the file pointed to by file descriptor number 10 with
   read permission at a location decided by the kernel.

mmap 0xdeadbeef 0x1000
 - Maps a new private+anonymous page with RWX permissions at a page boundary
   near 0xdeadbeef.
""",
)
@pwndbg.commands.OnlyWhenRunning
def mmap(addr, length, prot=7, flags=0x22, fd=-1, offset=0, quiet=False, force=False) -> None:
    try:
        prot_int = parse_str_or_int(prot, prot_str_to_val)
    except ValueError as e:
        print(message.error(f'Invalid protection value "{prot}": {e}'))
        return

    try:
        flag_int = parse_str_or_int(flags, flag_str_to_val)
    except ValueError as e:
        print(message.error(f'Invalid flags value "{flags}": {e}'))
        return

    aligned_addr = int(pwndbg.lib.memory.page_align(addr))
    if flag_int & flag_dict["MAP_FIXED"] != 0:
        # When using MAP_FIXED, it's only safe to call mmap(2) when the address
        # overlaps no other maps. We want to make sure that, unless the user
        # _really_ knows what they're doing, this call will be safe.
        #
        # Additionally, it's nice to highlight cases where the call is likely
        # to fail because the address is not properly aligned.
        addr = int(addr)
        if addr != aligned_addr and not quiet:
            print(
                message.warn(
                    f"""\
Address {addr:#x} is not properly aligned. Calling mmap with MAP_FIXED and an
unaligned address is likely to fail. Consider using the address {aligned_addr:#x}
instead.\
"""
                )
            )

        # Collision checking can get expensive for some combinations of number
        # of existing mappings and size of maps. If the user is using `--force`,
        # it's fair to assume they know what they're doing enough that we don't
        # need to bother them with any of this information, and get a nice
        # speedup as a bonus.
        if not force:
            page = pwndbg.lib.memory.Page(addr, int(length), 0, 0, pwndbg.aglib.arch.ptrsize)
            collisions = []
            vm = pwndbg.aglib.vmmap.get()

            # FIXME: The ends of the maps are sorted. We could bisect the array
            # in order to quickly reject all of the items we could never hit
            # (all of those such that `vm[i].end < page.start`).
            #
            # The target Python version as of the writing (3.8) does not support
            # `bissect.bissect_left(key=*)`, and cooking up our own
            # implementation feels overkill for something that could just be
            # fixed later with a version bump.
            for i in range(len(vm)):
                cand = vm[i]
                if cand.end > page.start and cand.start < page.end:
                    collisions.append(cand)
                if cand.start >= page.end:
                    # No more collisions are possible.
                    break

            if len(collisions) > 0:
                m = message.error
                print(
                    m(
                        f"""\
Trying to mmap with MAP_FIXED for an address range that collides with {len(collisions)}
existing range{"s" if len(collisions) > 1 else ""}:\
"""
                    )
                )
                for c in collisions:
                    print(m(f"    {c}"))
                print(
                    m(
                        """
This operation is destructive and will delete all of the listed mappings.\
"""
                    )
                )
                print(
                    m(
                        "Run this command again with `--force` if you still \
wish to proceed."
                    )
                )
                return

    elif int(addr) != aligned_addr and not quiet:
        # Highlight to the user that the address they've specified is likely to
        # be changed by the kernel.
        print(
            message.warn(
                f"""\
Address {addr:#x} is not properly aligned. It is likely to be changed to an
aligned address by the kernel automatically. If this is not desired, consider
using the address {aligned_addr:#x} instead.\
"""
            )
        )

    async def ctrl(ec: pwndbg.dbg_mod.ExecutionController):
        pointer = await pwndbg.aglib.shellcode.exec_syscall(
            ec,
            "SYS_mmap",
            int(pwndbg.lib.memory.page_align(addr)),
            int(length),
            prot_int,
            flag_int,
            int(fd),
            int(offset),
        )

        print(f"mmap syscall returned {pointer:#x}")

    pwndbg.dbg.selected_inferior().dispatch_execution_controller(ctrl)



from __future__ import annotations

import argparse
import binascii
import codecs
import os
import struct

import pwndbg
import pwndbg.aglib
import pwndbg.aglib.asm
import pwndbg.aglib.memory
import pwndbg.aglib.vmmap
import pwndbg.color.memory as mem_color
import pwndbg.commands
import pwndbg.enhance
import pwndbg.search
from pwndbg.color import message
from pwndbg.commands import CommandCategory

if pwndbg.dbg.is_gdblib_available():
    import gdb

saved: set[int] = set()


def print_search_hit(address: int) -> None:
    """Prints out a single search hit.

    Arguments:
        address: Address to print
    """
    if not address:
        return

    vmmap = pwndbg.aglib.vmmap.find(address)
    if vmmap:
        region = os.path.basename(vmmap.objfile)
    else:
        region = "[mapped]"

    region = region.ljust(15)

    region = mem_color.get(address, region)
    addr = mem_color.get(address)
    display = pwndbg.enhance.enhance(address)
    print(region, addr, display)


auto_save = pwndbg.config.add_param(
    "auto-save-search", False, 'automatically pass --save to "search" command'
)
parser = argparse.ArgumentParser(
    description="""Search memory for byte sequences, strings, pointers, and integer values.

By default search results are cached. If you want to cache all results, but only print a subset, use --trunc-out. If you want to cache only a subset of results, and print the results immediately, use --limit. The latter is specially useful if you're searching a huge section of memory.

""",
)
parser.add_argument(
    "-t",
    "--type",
    choices=["byte", "short", "word", "dword", "qword", "pointer", "string", "bytes", "asm"],
    help="Size of search target",
    default="bytes",
    type=str,
)
parser.add_argument(
    "-1",
    "--byte",
    dest="type",
    action="store_const",
    const="byte",
    help="Search for a 1-byte integer",
)
parser.add_argument(
    "-2",
    "--word",
    "--short",
    dest="type",
    action="store_const",
    const="word",
    help="Search for a 2-byte integer",
)
parser.add_argument(
    "-4",
    "--dword",
    dest="type",
    action="store_const",
    const="dword",
    help="Search for a 4-byte integer",
)
parser.add_argument(
    "-8",
    "--qword",
    dest="type",
    action="store_const",
    const="qword",
    help="Search for an 8-byte integer",
)
parser.add_argument(
    "-p",
    "--pointer",
    dest="type",
    action="store_const",
    const="pointer",
    help="Search for a pointer-width integer",
)
parser.add_argument(
    "--asm",
    dest="type",
    action="store_const",
    const="asm",
    help="Search for an assembly instruction",
)
parser.add_argument(
    "--asmbp", action="store_true", help="Set breakpoint for found assembly instruction"
)
parser.add_argument(
    "-x", "--hex", action="store_true", help="Target is a hex-encoded (for bytes/strings)"
)
parser.add_argument(
    "-e", "--executable", action="store_true", help="Search executable segments only"
)
parser.add_argument("-w", "--writable", action="store_true", help="Search writable segments only")
parser.add_argument(
    "-s",
    "--step",
    default=None,
    type=str,
    help="Step search address forward to next alignment after each hit (ex: 0x1000)",
)
parser.add_argument(
    "-l",
    "--limit",
    default=None,
    type=str,
    help="Max results before quitting the search. Differs from --trunc-out in that it will not save all search results before quitting",
)
parser.add_argument(
    "-a", "--aligned", default=None, type=str, help="Result must be aligned to this byte boundary"
)
parser.add_argument("value", type=str, help="Value to search for")
parser.add_argument(
    "mapping_name", type=str, nargs="?", default=None, help="Mapping to search [e.g. libc]"
)
parser.add_argument(
    "--save",
    action="store_true",
    default=None,
    help=f"Save results for further searches with --next. Default comes from config {auto_save.name!r}",
)
parser.add_argument(
    "--no-save", action="store_false", default=None, dest="save", help="Invert --save"
)
parser.add_argument(
    "-n",
    "--next",
    action="store_true",
    help="Search only locations returned by previous search with --save",
)
parser.add_argument(
    "--trunc-out",
    action="store_true",
    default=False,
    help="Truncate the output to 20 results. Differs from --limit in that it will first save all search results",
)


@pwndbg.commands.Command(parser, category=CommandCategory.MEMORY)
@pwndbg.commands.OnlyWhenRunning
def search(
    type,
    asmbp,
    hex,
    executable,
    writable,
    step,
    limit,
    aligned,
    value,
    mapping_name,
    save,
    next,
    trunc_out,
) -> None:
    global saved
    if next and not saved:
        print(
            "WARNING: cannot filter previous search results as they were empty. Performing new search saving results."
        )
        next = False
        save = True

    # Initialize is_pointer to track whether the search type is a pointer
    is_pointer = None
    # Adjust pointer sizes to the local architecture
    if type == "pointer":
        type = {4: "dword", 8: "qword"}[pwndbg.aglib.arch.ptrsize]
        is_pointer = True

    if save is None:
        save = bool(pwndbg.config.auto_save_search)

    if hex:
        try:
            value = codecs.decode(value, "hex")
        except binascii.Error as e:
            print(f"invalid input for type hex: {e}")
            return

    if step:
        step = pwndbg.commands.fix_int(step)

    if aligned:
        aligned = pwndbg.commands.fix_int(aligned)

    if limit:
        limit = pwndbg.commands.fix_int(limit)
    # Convert to an integer if needed, and pack to bytes
    if type not in ("string", "bytes", "asm"):
        value = pwndbg.commands.fix_int(value)
        value &= pwndbg.aglib.arch.ptrmask
        fmt = {"little": "<", "big": ">"}[pwndbg.aglib.arch.endian] + {
            "byte": "B",
            "short": "H",
            "word": "H",
            "dword": "L",
            "qword": "Q",
        }[type]

        try:
            value = struct.pack(fmt, value)
        except struct.error as e:
            print(f"invalid input for type {type}: {e}")
            return

    # Null-terminate strings
    elif type == "string":
        value = value.encode()
        value += b"\x00"

    elif type == "asm" or asmbp:
        value = pwndbg.aglib.asm.asm(value)

    # `pwndbg.search.search` expects a `bytes` object for its pattern. Convert the string pattern we
    # were given to a bytes object by encoding it as an UTF-8 byte sequence. This matches the behavior
    # we previously got by calling `gdb.Inferior.search_memory` with an `str`, since right about GDB
    # version 7.x or 8.x[1], as it uses a `Py_buffer` object populated with an `'s*'` pattern, which
    # has been encoding `str` object as a UTF-8 byte sequence since Python 3.1[2].
    #
    # [1]: https://sourceware.org/git/?p=binutils-gdb.git;a=blame;f=gdb/python/py-inferior.c;h=a1042ee72ac733091f7572bc04b072546d3c1519;hb=23c84db5b3cb4e8a0d555c76e1a0ab56dc8355f3
    # [2]: https://docs.python.org/3.1/c-api/arg.html#strings-and-buffers

    elif type == "bytes" and not hex:
        try:
            value = value.encode("utf-8")
        except UnicodeError as what:
            print(
                message.error(
                    f"Invalid pattern '{value}'. Patterns of type `bytes` must be encodable in UTF-8: {what}"
                )
            )
            return

    # Find the mappings that we're looking for
    mappings = pwndbg.aglib.vmmap.get()

    if mapping_name:
        mappings = [m for m in mappings if mapping_name in m.objfile]

    if not mappings:
        print(message.error(f"Could not find mapping {mapping_name!r}"))
        return

    # Output appropriate messages based on the detected search type for better clarity
    if is_pointer:
        print("Searching for a pointer-width integer: " + repr(value))
    elif type == "word" or type == "short":
        print("Searching for a 2-byte integer: " + repr(value))
    elif type == "dword":
        print("Searching for a 4-byte integer: " + repr(value))
    elif type == "qword":
        print("Searching for an 8-byte integer: " + repr(value))
    elif type == "string":
        print("Searching for string: " + repr(value))
    # If next is passed, only perform a manual search over previously saved addresses
    elif type == "asm" or asmbp:
        print("Searching for instruction (assembled value): " + repr(value))
    else:
        print("Searching for byte: " + repr(value))

    if next:
        val_len = len(value)
        new_saved = set()

        i = 0
        for addr in saved:
            try:
                val = pwndbg.aglib.memory.read(addr, val_len)
            except Exception:
                continue
            if val == value:
                new_saved.add(addr)
                if not trunc_out or i < 20:
                    print_search_hit(addr)
                i += 1

        print(f"Search found {i} items")
        saved = new_saved
        return

    # Prep the saved set if necessary
    if save:
        saved = set()

    # Perform the search
    i = 0
    for address in pwndbg.search.search(
        value,
        mappings=mappings,
        executable=executable,
        writable=writable,
        step=step,
        aligned=aligned,
        limit=limit,
    ):
        if save:
            saved.add(address)
        if asmbp:
            if pwndbg.dbg.is_gdblib_available():
                # set breakpoint on the instruction
                gdb.Breakpoint(f"*{address:#x}", temporary=False)
            else:
                print(
                    f"Breakpoints are not supported outside of GDB yet, would be set at {address:#x}"
                )

        if not trunc_out or i < 20:
            print_search_hit(address)
        i += 1



"""
Enables callbacks into functions to be automatically invoked
when various events occur to the debuggee (e.g. STOP on SIGINT)
by using a decorator.
"""

from __future__ import annotations

import functools
import sys
from collections import defaultdict
from collections import deque
from collections.abc import Callable
from functools import partial
from functools import wraps
from typing import Any
from typing import TypeVar

import gdb
from typing_extensions import ParamSpec

import pwndbg
import pwndbg.lib.config
from pwndbg import config
from pwndbg.color import message
from pwndbg.dbg_mod import EventHandlerPriority

DISABLED = "disabled"
DISABLED_DEADLOCK = "disabled-deadlock"
ENABLED = "enabled"

gdb_workaround_stop_event = config.add_param(
    "gdb-workaround-stop-event",
    DISABLED,
    "asynchronous stop events to improve 'commands' functionality",
    help_docstring=f"""
Note that this may cause unexpected behavior with Pwndbg or gdb.execute.

Values explained:

+ `{DISABLED}` - Disable the workaround (default).
+ `{DISABLED_DEADLOCK}` - Disable only deadlock detection; deadlocks may still occur.
+ `{ENABLED}` - Enable asynchronous stop events; gdb.execute may behave unexpectedly (asynchronously).
    """,
    param_class=pwndbg.lib.config.PARAM_ENUM,
    enum_sequence=[DISABLED, DISABLED_DEADLOCK, ENABLED],
)

P = ParamSpec("P")
T = TypeVar("T")


# There is no GDB way to get a notification when the binary itself
# is loaded from disk, by the operating system, before absolutely
# anything happens
#
# However, we get an Objfile event when the binary is loaded, before
# its entry point is invoked.
#
# We also get an Objfile event when we load up GDB, so we need
# to detect when the binary is running or not.
#
# Additionally, when attaching to a process running under QEMU, the
# very first event which is fired is a 'stop' event.  We need to
# capture this so that we can fire off all of the 'start' events first.
class StartEvent:
    def __init__(self) -> None:
        self.registered: list[Callable[..., Any]] = []
        self.running = False

    def connect(self, function: Callable[..., Any]) -> None:
        if function not in self.registered:
            self.registered.append(function)

    def disconnect(self, function: Callable[..., Any]) -> None:
        if function in self.registered:
            self.registered.remove(function)

    def on_new_objfile(self) -> None:
        if self.running or not gdb.selected_thread():
            return

        self.running = True

        for function in self.registered:
            function()

    def on_exited(self) -> None:
        self.running = False

    def on_stop(self) -> None:
        self.on_new_objfile()


# Monkeypatching GDB hihi
gdb.events.start = StartEvent()
gdb.events.suspend_all = object()


def _is_safe_event_packet():
    try:
        gdb.selected_frame()
    except gdb.error as e:
        if "Remote 'g' packet reply is too long" in str(e):
            return False
    return True


def _is_safe_event_thread():
    try:
        gdb.newest_frame()
    except gdb.error as e:
        if "Selected thread is running" in str(e):
            return False
    return True


queued_events: deque[Callable[..., Any]] = deque()
executing_event = False
workaround_thread_conn = None


def _update_start_event_state(event_type: Any):
    """
    Update the state of the StartEvent appropriately
    (we emulate this event so we need to set it properly)
    """
    # Implement our custom event gdb.events.start!
    if event_type == gdb.events.stop:
        queued_events.append(gdb.events.start.on_stop)
    elif event_type == gdb.events.new_objfile:
        queued_events.append(gdb.events.start.on_new_objfile)
    elif event_type == gdb.events.exited:
        gdb.events.start.on_exited()


def _detect_deadlock():
    if not executing_event:
        # Not executing an event inside another event, so no deadlock
        return

    if gdb_workaround_stop_event == DISABLED_DEADLOCK:
        # Skip deadlock detection because this option disables it
        return

    print(message.error("DEADLOCK DETECTED..."))
    print(
        message.error(
            f"""The deadlock issue is likely caused by using 'commands[\\n]{
                message.hint("continue")
            }[\\n]end'.

To address this, you have three options:

1. Avoid using '{message.hint("commands")}'. Instead, rewrite it as a Python script. For example:
   {
                message.hint('''
   # Read more at: https://github.com/pwndbg/pwndbg/issues/425#issuecomment-892302716
   class Bp(gdb.Breakpoint):
       def stop(self):
           print("Breakpoint hit!")
           return False  # False = continue to next breakpoint, True = stop inferior

   Bp("main")
   ''')
            }

2. Replace '{message.hint("continue")}' with '{
                message.hint('pi gdb.execute("continue")')
            }' and use '{message.hint("set gdb-workaround-stop-event disabled-deadlock")}'.
   This change reduces the likelihood of deadlocks, while preserving pwndbg functionality.

3. Run '{message.hint("set gdb-workaround-stop-event enabled")}', allowing you to keep '{
                message.hint("continue")
            }' as is.
   However, this setting may cause pwndbg or gdb.execute to behave asynchronously/unpredictably.
"""
        )
    )
    sys.exit(1)


def wrap_safe_event_handler(event_handler: Callable[P, T], event_type: Any) -> Callable[P, T]:
    """
    Wraps an event handler to ensure it is only executed when the event is safe.
    Invalid events are queued and executed later when safe.

    Note: Avoid using `gdb.post_event` because of another bug in gdbserver
    where the `gdb.newest_frame` function may not work properly.

    Workaround to fix bug in gdbserver (gdb.events.new_objfile): https://github.com/pwndbg/pwndbg/issues/2576
    Workaround to fix bug in gdb (gdb.events.stop): https://github.com/pwndbg/pwndbg/issues/425
    """

    def _loop_until_thread_ok():
        global queued_events

        if not queued_events:
            return

        if not _is_safe_event_thread():
            gdb.post_event(_loop_until_thread_ok)
            return

        while queued_events:
            queued_events.popleft()()

    @wraps(event_handler)
    def _inner_handler(*a: P.args, **kw: P.kwargs):
        global queued_events, executing_event

        if event_type == gdb.events.start:
            # SKIP our custom event from this wrapper...
            event_handler(*a, **kw)
            return

        _detect_deadlock()
        _update_start_event_state(event_type)

        queued_events.append(lambda: event_handler(*a, **kw))
        if event_type == gdb.events.new_objfile and not _is_safe_event_packet():
            # Workaround to issue with gdbserver - Remote 'g' packet reply is too long
            # https://github.com/pwndbg/pwndbg/issues/2576
            gdb.post_event(_loop_until_thread_ok)
            return
        if event_type == gdb.events.stop:
            # Workaround to issue with gdb `commands \n continue \n end` - Selected thread is running
            # https://github.com/pwndbg/pwndbg/issues/425
            if gdb_workaround_stop_event == ENABLED:
                gdb.post_event(_loop_until_thread_ok)
                return

            executing_event = True
            gdb.execute("", to_string=True)  # Trigger bug in gdb, it is like 'yield'
            executing_event = False
        elif event_type in (gdb.events.cont, gdb.events.new_thread):
            # Workaround for crash in gdb when used: `target extended-remote` + `attach`
            # https://github.com/pwndbg/pwndbg/issues/3231
            global workaround_thread_conn
            conn = gdb.selected_inferior().connection
            if (
                isinstance(conn, gdb.RemoteTargetConnection)
                and conn.type == "extended-remote"
                and conn.is_valid()
                and workaround_thread_conn != conn
            ):
                gdb.selected_inferior().threads()[0].switch()
                workaround_thread_conn = conn

        while queued_events:
            queued_events.popleft()()

    return _inner_handler


# In order to support reloading, we must be able to re-fire
# all 'objfile' and 'stop' events.
registered: dict[gdb.EventRegistry[Any], dict[EventHandlerPriority, list[Callable[..., None]]]] = {
    gdb.events.exited: {},
    gdb.events.cont: {},
    gdb.events.new_objfile: {},
    gdb.events.stop: {},
    gdb.events.start: {},
    gdb.events.new_thread: {},
    gdb.events.before_prompt: {},  # The real event might not exist, but we wrap it
    gdb.events.memory_changed: {},
    gdb.events.register_changed: {},
}

# Keys are gdb.events.*
paused = defaultdict(bool)


def pause(event_registry: gdb.EventRegistry[Any]) -> None:
    paused[event_registry] = True


def unpause(event_registry: gdb.EventRegistry[Any]) -> None:
    paused[event_registry] = False


def event_handler_factory(
    event_registry: gdb.EventRegistry[Any], event_name: str, priority: EventHandlerPriority
) -> Callable[[Callable[..., None]], Callable[..., None]]:
    """
    Essentially the implementation of pwndbg.dbg_mod.gdb.event_handler().

    Takes a gdb.EventRegistry because that contains some events that pwndbg.dbg_mod.EventType
    doesn't.

    Read the first few paragraphs of:
    https://sourceware.org/gdb/current/onlinedocs/gdb.html/Events-In-Python.html#Events-In-Python
    """

    def decorator(fn: Callable[..., None]) -> Callable[..., None]:
        # This is only executed once.

        if pwndbg.config.dev_debug_events:
            print("Connecting", fn.__name__, event_name)

        should_connect: bool = False

        if not registered[event_registry]:
            # We've never seen this type of event before.
            should_connect = True

        # Wrap the function for dev instrumentation
        @functools.wraps(fn)
        def _dev_wrapper(*a, **kw) -> None:
            if pwndbg.config.dev_debug_events:
                sys.stdout.write(
                    f"{event_name} ({priority.name}) {fn.__module__}.{fn.__qualname__}\n"
                )

            return fn(*a, **kw)

        # Register this event handler with us.
        registered[event_registry].setdefault(priority, []).append(_dev_wrapper)

        if should_connect:
            # We actually need to tell gdb to run `invoke_event` when
            # the appropriate event is invoked. Then inside `invoke_event`
            # we will call all `registered` functions.

            wrapped_invoke_event = partial(invoke_event, event_registry)
            wrapped_invoke_event = wrap_safe_event_handler(wrapped_invoke_event, event_registry)

            event_registry.connect(wrapped_invoke_event)

        return _dev_wrapper

    return decorator


def log_objfiles(ofile: gdb.NewObjFileEvent | None = None) -> None:
    if not (pwndbg.config.dev_debug_events and ofile):
        return None

    name = ofile.new_objfile.filename

    print(f"objfile: {name!r}")
    gdb.execute("info sharedlibrary")
    return None


gdb.events.new_objfile.connect(log_objfiles)


def invoke_event(event_registry: gdb.EventRegistry[Any], event_data: Any = None) -> None:
    """
    Invoke all registered handlers for a certain event_registry.
    event_data may be None if we manually ran this function.
    """
    if paused[event_registry] or paused[gdb.events.suspend_all]:
        return

    handlers = registered.get(event_registry)
    if handlers is None:
        return

    try:
        # Run all the event handlers for this event from lowest to highest priority.
        for prio in EventHandlerPriority:
            for func in handlers.get(prio, []):
                func()
    except Exception as e:
        import pwndbg.exception

        pwndbg.exception.handle()
        raise e


def after_reload(fire_start: bool = True) -> None:
    if gdb.selected_inferior().pid:
        invoke_event(gdb.events.stop)
        if fire_start:
            invoke_event(gdb.events.start)
        invoke_event(gdb.events.new_objfile)
        invoke_event(gdb.events.before_prompt)


def on_reload() -> None:
    for functions in registered.values():
        functions.clear()
